"""
Unit tests for the sdk implementations.
"""

import logging
import os.path
import tempfile
import unittest
from datetime import datetime, timedelta

import numpy as np

import finam as fm
from finam import (
    Adapter,
    CallbackInput,
    CallbackOutput,
    ComponentStatus,
    Composition,
    EsriGrid,
    FinamDataError,
    FinamLogError,
    FinamMetaDataError,
    FinamNoDataError,
    FinamStaticDataError,
    FinamStatusError,
    FinamTimeError,
    Info,
    Input,
    NoGrid,
    Output,
    TimeComponent,
)
from finam.sdk.component import IOList


class MockupAdapter(Adapter):
    def __init__(self):
        super().__init__()

    def _get_data(self, time):
        return time


class MockupComponent(TimeComponent):
    def __init__(self):
        super().__init__()
        self._time = datetime(2000, 1, 1)

    def _next_time(self):
        return self.time + timedelta(days=1)

    def _initialize(self):
        self.status = ComponentStatus.FAILED


class MockupComponentIO(TimeComponent):
    def __init__(self):
        super().__init__()
        self._time = datetime(2000, 1, 1)

    def _next_time(self):
        return self.time + timedelta(days=1)

    def _initialize(self):
        self.inputs.add(name="Input")
        self.outputs.add(name="Output")


class MockupComponentIONameConflict(TimeComponent):
    def __init__(self):
        super().__init__()
        self._time = datetime(2000, 1, 1)

    def _next_time(self):
        return self.time + timedelta(days=1)

    def _initialize(self):
        self.inputs.add(name="IO")
        self.outputs.add(name="IO")


class TestComponent(unittest.TestCase):
    def test_component_status(self):
        component = MockupComponent()

        self.assertEqual(component.time, datetime(2000, 1, 1))
        self.assertEqual(component.status, ComponentStatus.CREATED)

        with self.assertRaises(FinamStatusError):
            Composition([component])

    def test_connect_helper(self):
        component = MockupComponentIO()
        composition = Composition([component])

        component.create_connector()

        self.assertTrue(component.connector is not None)
        self.assertEqual(component.connector._inputs, component.inputs)
        self.assertEqual(component.connector._outputs, component.outputs)

        component.try_connect(None)

        self.assertEqual(component.status, ComponentStatus.CONNECTING_IDLE)

    def test_simple_io_access(self):
        comp_ok = MockupComponentIO()
        comp_fail = MockupComponentIONameConflict()

        comp_ok.initialize()
        comp_fail.initialize()

        inp = comp_ok["Input"]
        out = comp_ok["Output"]

        self.assertTrue(isinstance(inp, fm.IInput))
        self.assertTrue(isinstance(out, fm.IOutput))

        with self.assertRaises(KeyError):
            _inp = comp_ok["abc"]

        with self.assertRaises(KeyError):
            _inp = comp_fail["IO"]

    def test_simple_io_not_initialized(self):
        comp_ok = MockupComponentIO()

        with self.assertRaises(KeyError):
            _inp = comp_ok["Input"]
        with self.assertRaises(KeyError):
            _inp = comp_ok["Output"]


class TestChaining(unittest.TestCase):
    def test_chaining(self):
        adapter1 = MockupAdapter()
        adapter2 = MockupAdapter()
        adapter3 = MockupAdapter()

        adapter1 >> adapter2 >> adapter3

        self.assertEqual(adapter1.targets, [adapter2])
        self.assertEqual(adapter2.targets, [adapter3])

        self.assertEqual(adapter2.source, adapter1)
        self.assertEqual(adapter3.source, adapter2)

    def test_multiple_sources(self):
        adapter1 = MockupAdapter()
        adapter2 = MockupAdapter()
        adapter3 = MockupAdapter()

        adapter1 >> adapter3

        with self.assertRaises(ValueError) as context:
            adapter2 >> adapter3

        self.assertTrue("Source of input is already set!" in str(context.exception))


class TestOutput(unittest.TestCase):
    def test_push_notify(self):
        counter = 0
        t = datetime(2000, 1, 1)
        info = Info(time=t, grid=NoGrid(), meta={"test": 0})
        wrong_info = Info(time=t, grid=EsriGrid(2, 2), meta={"test": 5})

        def callback(_clr, _time):
            nonlocal counter
            counter += 1

        out = Output(name="Output")
        inp = CallbackInput(callback=callback, name="Callback")

        out >> inp

        with self.assertRaises(FinamNoDataError):
            out.get_data(t, None)

        out.push_info(info)
        out._connected_inputs = {inp: None}

        with self.assertRaises(FinamNoDataError):
            out.push_data(100, t)

        out._out_infos_exchanged = 1

        out.get_info(info)

        with self.assertRaises(FinamMetaDataError):
            out.get_info(wrong_info)

        out._output_info.time = None
        with self.assertRaises(FinamMetaDataError):
            out.get_info(Info(time=None, grid=NoGrid()))

        out._output_info.time = t

        out.push_data(100, t)

        self.assertTrue(inp.has_source)
        self.assertTrue(out.has_targets)
        self.assertEqual(out.get_info(info), info)
        self.assertEqual(out.get_data(t, None), 100)
        self.assertEqual(inp.exchange_info(info), info)
        self.assertEqual(inp.pull_data(t), 100)
        self.assertEqual(counter, 1)

    def test_know_targets(self):
        out = Output(name="Output")
        in1 = Input(name="In1")
        in2 = Input(name="In1")

        ada = fm.adapters.Scale(2.0)

        out >> ada
        ada >> in1
        ada >> in2

        in1.ping()
        in2.ping()

        self.assertEqual(out._connected_inputs, {in1: None, in2: None})

        with self.assertRaises(ValueError):
            in1.ping()

    def test_know_targets_adapter(self):
        out = Output(name="Output")
        in1 = Input(name="In1")
        in2 = Input(name="In1")

        ada = fm.adapters.LinearTime()

        out >> ada
        ada >> in1
        ada >> in2

        in1.ping()
        in2.ping()

        self.assertEqual(out._connected_inputs, {ada: None})

    def test_cache_data(self):
        t = datetime(2000, 1, 1)
        info = Info(time=t, grid=NoGrid())

        out = Output(name="Output")
        inp = Input(name="Input")

        out >> inp

        inp.ping()

        out.push_info(info)
        inp.exchange_info(info)

        self.assertEqual(out._connected_inputs, {inp: None})

        out.push_data(1, t)
        self.assertEqual(out.data, [(t, 1)])

        with self.assertRaises(FinamTimeError):
            data = inp.pull_data(t - timedelta(hours=1))
        with self.assertRaises(FinamTimeError):
            data = inp.pull_data(t + timedelta(hours=1))

        data = inp.pull_data(t)
        self.assertEqual(out.data, [(t, 1)])
        self.assertEqual(data, 1)

        t2 = t + timedelta(days=1)
        out.push_data(2, t2)
        self.assertEqual(out.data, [(t, 1), (t2, 2)])

        with self.assertRaises(FinamTimeError):
            data = inp.pull_data(t - timedelta(hours=1))
        with self.assertRaises(FinamTimeError):
            data = inp.pull_data(t2 + timedelta(hours=1))

        self.assertEqual(inp.pull_data(t), 1)
        self.assertEqual(inp.pull_data(t + timedelta(hours=0)), 1)
        self.assertEqual(inp.pull_data(t + timedelta(hours=11)), 1)
        self.assertEqual(inp.pull_data(t + timedelta(hours=12)), 2)

        self.assertEqual(len(out.data), 2)
        self.assertEqual(inp.pull_data(t2), 2)
        self.assertEqual(len(out.data), 1)

        out.push_data(3, t + timedelta(days=2))
        out.push_data(4, t + timedelta(days=3))
        out.push_data(5, t + timedelta(days=4))

        self.assertEqual(len(out.data), 4)
        self.assertEqual(inp.pull_data(t + timedelta(days=2)), 3)
        self.assertEqual(len(out.data), 3)

    def test_cache_data_multi(self):
        t = datetime(2000, 1, 1)
        info = Info(time=t, grid=NoGrid())

        out = Output(name="Output")
        in1 = Input(name="Input")
        in2 = Input(name="Input")

        out >> in1
        out >> in2

        in1.ping()
        in2.ping()

        out.push_info(info)
        in1.exchange_info(info)
        in2.exchange_info(info)

        for i in range(10):
            out.push_data(i + 1, t + timedelta(days=i))

        self.assertEqual(len(out.data), 10)

        in1.pull_data(datetime(2000, 1, 3))
        self.assertEqual(len(out.data), 10)

        in2.pull_data(datetime(2000, 1, 1))
        self.assertEqual(len(out.data), 10)

        in2.pull_data(datetime(2000, 1, 2))
        self.assertEqual(len(out.data), 9)

        in2.pull_data(datetime(2000, 1, 8))
        self.assertEqual(len(out.data), 8)

        in1.pull_data(datetime(2000, 1, 8))
        self.assertEqual(len(out.data), 3)

        in1.pull_data(datetime(2000, 1, 10))
        self.assertEqual(len(out.data), 3)
        in2.pull_data(datetime(2000, 1, 10))
        self.assertEqual(len(out.data), 1)

    def test_push_static(self):
        t = datetime(2000, 1, 1)
        info = Info(time=t, grid=NoGrid())

        out = Output(name="Output", static=True)
        in1 = Input(name="Input")

        out >> in1

        in1.ping()

        out.push_info(info)
        in1.exchange_info(info)

        out.push_data(0, None)

        with self.assertRaises(FinamStaticDataError):
            out.push_data(0, None)

    def test_data_copied(self):
        t = datetime(2000, 1, 1)
        info = Info(time=t, grid=fm.UniformGrid((1, 1)))

        out = Output(name="Output")
        in1 = Input(name="Input")

        out >> in1

        in1.ping()

        out.push_info(info)
        in1.exchange_info(info)

        in_data = fm.data.full(0.0, info)
        out.push_data(in_data, t)
        with self.assertRaises(FinamDataError):
            out.push_data(in_data, t)

    def test_data_copied_units(self):
        t = datetime(2000, 1, 1)
        info1 = Info(time=t, grid=fm.UniformGrid((1, 1)), units="m")
        info2 = Info(time=t, grid=fm.UniformGrid((1, 1)), units="km")

        out = Output(name="Output")
        in1 = Input(name="Input")

        out >> in1

        in1.ping()

        out.push_info(info1)
        in1.exchange_info(info2)

        in_data = fm.data.full(0.0, info1)
        out.push_data(in_data, t)
        out_data = in1.pull_data(t, in1)

        self.assertEqual(out_data[0, 0, 0], 0.0 * fm.UNITS("km"))
        in_data[0, 0] = 1.0 * fm.UNITS("m")
        self.assertEqual(out_data[0, 0, 0], 0.0 * fm.UNITS("km"))

    def test_memory_limit(self):
        t = datetime(2000, 1, 1)
        info = Info(time=t, grid=fm.UniformGrid((100, 100)))

        with tempfile.TemporaryDirectory() as td:
            out = Output(name="Output")
            out.memory_limit = 0
            out.memory_location = td
            oid = id(out)

            in1 = Input(name="Input")

            out >> in1

            in1.ping()

            out.push_info(info)
            in1.exchange_info(info)

            in_data = fm.data.full(0.0, info)
            out.push_data(np.copy(in_data), datetime(2000, 1, 1))
            out.push_data(np.copy(in_data), datetime(2000, 1, 2))

            self.assertTrue(os.path.isfile(os.path.join(td, f"{oid}-{0}.npy")))
            self.assertTrue(os.path.isfile(os.path.join(td, f"{oid}-{1}.npy")))

            data = in1.pull_data(datetime(2000, 1, 2), in1)

            np.testing.assert_allclose(data.magnitude, in_data.magnitude)
            self.assertEqual(data.units, in_data.units)
            self.assertEqual(data.units, info.units)

            self.assertFalse(os.path.isfile(os.path.join(td, f"{oid}-{0}.npy")))
            self.assertTrue(os.path.isfile(os.path.join(td, f"{oid}-{1}.npy")))

            out.finalize()

            self.assertFalse(os.path.isfile(os.path.join(td, f"{oid}-{0}.npy")))
            self.assertFalse(os.path.isfile(os.path.join(td, f"{oid}-{1}.npy")))


class TestInput(unittest.TestCase):
    def test_fail_set_source(self):
        time = datetime(2000, 1, 1)
        inp = Input(name="In", time=time, grid=NoGrid())
        outp = Output(name="Out", time=time, grid=NoGrid())

        with self.assertRaises(ValueError):
            inp.source = 0

        inp.source = outp

        with self.assertRaises(ValueError):
            inp.source = outp

    def test_fail_accept(self):
        t = datetime(2000, 1, 1)
        info = Info(time=t, grid=NoGrid())
        info2 = Info(time=t, grid=NoGrid(), units="m")

        out = Output(name="Output", static=True)
        in1 = Input(name="Input", static=True)

        out >> in1

        in1.ping()

        out.push_info(info)

        with self.assertRaises(FinamMetaDataError):
            in1.exchange_info(info2)

        in1.exchange_info(info)

        with self.assertRaises(FinamMetaDataError):
            in1.exchange_info(info)

    def test_pull_static(self):
        t = datetime(2000, 1, 1)
        info = Info(time=t, grid=NoGrid())

        out = Output(name="Output", static=True)
        in1 = Input(name="Input", static=True)

        out >> in1

        in1.ping()

        out.push_info(info)
        in1.exchange_info(info)

        out.push_data(0, None)
        data = in1.pull_data(None)

        self.assertTrue(fm.data.has_time_axis(data, info.grid))

        data_2 = in1.pull_data(None)

        self.assertEqual(data, in1._cached_data)
        self.assertEqual(data, data_2)

    def test_pull_dynamic_time(self):
        t = datetime(2000, 1, 1)
        info = Info(time=t, grid=NoGrid())

        out = Output(name="Output", static=True)
        in1 = Input(name="Input")

        out >> in1

        in1.ping()

        out.push_info(info)
        in1.exchange_info(info)

        out.push_data(0, None)
        data = in1.pull_data(t)

        self.assertTrue(fm.data.has_time_axis(data, info.grid))


class TestCallbackInput(unittest.TestCase):
    def test_callback_input(self):
        caller = None
        counter = 0
        t = datetime(2000, 1, 1)

        def callback(clr, _time):
            nonlocal caller
            nonlocal counter
            caller = clr
            counter += 1

        inp = CallbackInput(callback=callback, name="callback")
        out = Output(name="Out")

        out >> inp

        inp.source_updated(t)

        self.assertEqual(caller, inp)
        self.assertEqual(counter, 1)


class TestCallbackOutput(unittest.TestCase):
    def test_callback_output(self):
        caller = None
        counter = 0
        t = datetime(2000, 1, 1)

        def callback(clr, _time):
            nonlocal caller
            nonlocal counter
            caller = clr
            counter += 1
            return 42 if counter == 1 else None

        out = CallbackOutput(callback=callback, name="callback")
        inp = Input(name="input", time=t, grid=NoGrid())

        out >> inp
        inp.ping()

        with self.assertRaises(NotImplementedError):
            _data = out.push_data(0, t)

        with self.assertRaises(ValueError):
            _data = out.get_data(0, None)

        with self.assertRaises(FinamNoDataError):
            _data = out.get_data(t, None)

        out._output_info = Info(time=t, grid=NoGrid())

        with self.assertRaises(FinamNoDataError):
            _data = out.get_data(t, None)

        out._out_infos_exchanged = 1

        data = out.get_data(t, None)

        self.assertEqual(data, 42)
        self.assertEqual(caller, out)
        self.assertEqual(counter, 1)

        with self.assertRaises(FinamNoDataError):
            _data = out.get_data(t, None)


class TestIOList(unittest.TestCase):
    def test_io_list(self):
        inp = Input("test1")
        out = Output("test2")
        inp_list = IOList(None, "INPUT")
        out_list = IOList(None, "OUTPUT")

        # io setting
        inp_list.add(inp)
        out_list["test2"] = out
        inp_list.add(name="test0")

        # wrong type
        with self.assertRaises(ValueError):
            inp_list.add(out)
        with self.assertRaises(ValueError):
            out_list["test1"] = inp

        # no logger
        with self.assertRaises(FinamLogError):
            inp_list.set_logger(None)

        # already present
        with self.assertRaises(ValueError):
            inp_list["test1"] = inp
        with self.assertRaises(ValueError):
            out_list.add(out)

        # wrong name
        with self.assertRaises(ValueError):
            inp_list["test2"] = inp

        # make frozen
        inp_list.frozen = True
        out_list.frozen = True
        with self.assertRaises(ValueError):
            inp_list["test1"] = inp
        with self.assertRaises(ValueError):
            out_list.add(name="test3")

    def test_io_list_fail(self):
        comp = MockupComponent()
        inp_list = IOList(None, "INPUT")

        with self.assertRaises(KeyError):
            inp_list["In"]

        inp_list.owner = comp
        with self.assertRaises(KeyError):
            inp_list["In"]

        comp.initialize()
        with self.assertRaises(KeyError):
            inp_list["In"]


class TestComponentFails(unittest.TestCase):
    def test_try_connect_fail(self):
        comp = MockupComponent()
        with self.assertRaises(FinamStatusError):
            comp.try_connect(None)

    def test_time_fail(self):
        comp = MockupComponent()
        comp._time = 0
        with self.assertRaises(ValueError):
            _t = comp.time

        with self.assertRaises(ValueError):
            comp.time = 0

    def test_get_slot_fail(self):
        comp = MockupComponentIONameConflict()
        comp.initialize()

        with self.assertRaises(KeyError):
            _ = comp["IO"]

        comp = MockupComponentIO()

        with self.assertRaises(KeyError):
            _ = comp["Input"]

        with self.assertRaises(KeyError):
            _ = comp["Output"]

        comp.initialize()
        _ = comp["Input"]
        _ = comp["Output"]

        with self.assertRaises(KeyError):
            _ = comp["x"]

        with self.assertRaises(KeyError):
            _ = comp["x"]


class TestAdapter(unittest.TestCase):
    def test_adapter(self):
        adapter = MockupAdapter()
        adapter.push_data(1, datetime(2000, 1, 1))

        with self.assertRaises(ValueError):
            adapter.push_data(1, 0)

        with self.assertRaises(ValueError):
            adapter.source_updated(0)

        with self.assertRaises(FinamMetaDataError):
            adapter.exchange_info(None)

        with self.assertRaises(FinamMetaDataError):
            adapter.exchange_info(0)

        with self.assertRaises(FinamLogError):
            adapter.source = 0

        self.assertEqual(adapter.info, None)

    def test_adapter_static(self):
        adapter = MockupAdapter()
        self.assertFalse(adapter.is_static)


class TestIOFails(unittest.TestCase):
    def test_input_output_fail(self):
        t = datetime(2000, 1, 1)

        with self.assertRaises(ValueError):
            _in = Input(name=None)
        with self.assertRaises(ValueError):
            _in = Input(name="In", info=Info(time=t, grid=NoGrid()), units="m")

        with self.assertRaises(ValueError):
            _in = Output(name=None)
        with self.assertRaises(ValueError):
            _in = Output(name="In", info=Info(time=t, grid=NoGrid()), units="m")

        inp = Input(name="In", time=t, grid=NoGrid())
        out = Output(name="Out", time=t, grid=NoGrid())
        out >> inp
        with self.assertRaises(ValueError):
            inp.pull_data(0)

        inp = Input(name="In", time=t, grid=NoGrid())
        out = Output(name="Out", time=t, grid=NoGrid())
        out >> inp
        inp._in_info_exchanged = True
        with self.assertRaises(FinamMetaDataError):
            inp.exchange_info(info=Info(time=t, grid=NoGrid()))

        inp = Input(name="In")
        out = Output(name="Out", time=t, grid=NoGrid())
        out >> inp
        with self.assertRaises(FinamMetaDataError):
            inp.exchange_info(info=None)

        inp = Input(name="In", time=t, grid=NoGrid())
        out = Output(name="Out", time=t, grid=NoGrid())
        out >> inp
        with self.assertRaises(FinamMetaDataError):
            inp.exchange_info(info=Info(time=t, grid=NoGrid()))

        inp = Input(name="In")
        out = Output(name="Out", time=t, grid=NoGrid())
        out >> inp
        with self.assertRaises(FinamMetaDataError):
            inp.exchange_info(info=100)

        out = Output(name="Out", time=t, grid=NoGrid())
        with self.assertRaises(ValueError):
            out.add_target(0)

        out.push_data(0, datetime(2000, 1, 1))

        inp = Input(name="In")
        out >> inp
        with self.assertRaises(ValueError):
            out.push_data(0, 0)

        with self.assertRaises(FinamMetaDataError):
            out.push_info(0)

        with self.assertRaises(ValueError):
            out.notify_targets(0)

        with self.assertRaises(ValueError):
            out.get_data(0, None)

        out = Output(name="Out")
        with self.assertRaises(FinamMetaDataError):
            out.push_info(0)

        with self.assertRaises(FinamNoDataError):
            out.get_info(Info(time=t, grid=NoGrid()))

        out.push_info(Info(time=t, grid=None, units=None))
        with self.assertRaises(FinamMetaDataError):
            out.get_info(Info(time=t, grid=None))

        with self.assertRaises(FinamMetaDataError):
            out.get_info(Info(time=t, grid=NoGrid(), units=None))

        with self.assertRaises(FinamMetaDataError):
            out.get_info(Info(time=t, grid=NoGrid(), units=None))

    def test_callback_input_fail(self):
        inp = CallbackInput(callback=lambda t: t, name="In")
        out = Output(name="Out")

        out >> inp

        with self.assertRaises(ValueError):
            inp.source_updated(0)


class NotImplComponent(TimeComponent):
    def __init__(self):
        super().__init__()
        self._time = datetime(2000, 1, 1)

    def _next_time(self):
        return self.time + timedelta(days=1)


class NotImplAdapter(Adapter):
    def __init__(self):
        super().__init__()


class TestNotImplemented(unittest.TestCase):
    def test_comp_not_implemented(self):
        comp = NotImplComponent()

        with self.assertRaises(NotImplementedError):
            comp.initialize()
        with self.assertRaises(NotImplementedError):
            comp._initialize()

        with self.assertRaises(NotImplementedError):
            comp.connect(datetime(2000, 1, 1))
        with self.assertRaises(NotImplementedError):
            comp._connect(datetime(2000, 1, 1))

        # check that the debug log for not implementing _validate is there
        with self.assertLogs(level=logging.DEBUG) as captured:
            comp.validate()
        self.assertEqual(len(captured.records), 2)
        self.assertEqual(captured.records[0].levelno, logging.DEBUG)
        self.assertEqual(captured.records[1].levelno, logging.DEBUG)
        with self.assertLogs(level=logging.DEBUG) as captured:
            comp._validate()
        self.assertEqual(len(captured.records), 1)
        self.assertEqual(captured.records[0].levelno, logging.DEBUG)

        with self.assertRaises(NotImplementedError):
            comp.update()
        with self.assertRaises(NotImplementedError):
            comp._update()

        # check that the debug log for not implementing _finalize is there
        with self.assertLogs(level=logging.DEBUG) as captured:
            comp.finalize()
        self.assertEqual(len(captured.records), 2)
        self.assertEqual(captured.records[0].levelno, logging.DEBUG)
        self.assertEqual(captured.records[1].levelno, logging.DEBUG)
        with self.assertLogs(level=logging.DEBUG) as captured:
            comp._finalize()
        self.assertEqual(len(captured.records), 1)
        self.assertEqual(captured.records[0].levelno, logging.DEBUG)

    def test_adapter_not_implemented(self):
        adapter = NotImplAdapter()
        with self.assertRaises(NotImplementedError):
            adapter.get_data(datetime(2000, 1, 1), None)
        with self.assertRaises(NotImplementedError):
            adapter._get_data(datetime(2000, 1, 1), None)


class TestRename(unittest.TestCase):
    def test_rename_component(self):
        comp = fm.components.SimplexNoise().with_name("CompA")
        self.assertEqual("CompA", comp.name)

    def test_rename_adapter(self):
        ada = fm.adapters.Scale(1.0).with_name("Scale-1")
        self.assertEqual("Scale-1", ada.name)


if __name__ == "__main__":
    unittest.main()



"""Helper functions to inspect FINAM entities."""
from ..data.tools import Info
from ..errors import FinamNoDataError
from ..interfaces import IAdapter, IComponent, IInput, IOutput, ITimeComponent

DEFAULT_INDENT = "  "


def inspect(obj):
    """Prints information about a component, adapter or input/output slot."""
    if isinstance(obj, IComponent):
        return _inspect_component(obj)
    if isinstance(obj, IAdapter):
        return _inspect_slot(obj)
    if isinstance(obj, (IInput, IOutput)):
        return _inspect_slot(obj)
    if isinstance(obj, Info):
        return _inspect_info(obj)
    raise ValueError(
        f"Can only inspect FINAM components, adapters and i/o slots. Got {obj.__class__.__name__}."
    )


def _inspect_component(component, indent=""):
    s = indent + component.name
    if isinstance(component, ITimeComponent):
        s += f"\n{indent}{DEFAULT_INDENT}time: {component.time}"
    if len(component.inputs) > 0:
        s += f"\n{indent}{DEFAULT_INDENT}Inputs"
        for _name, inp in component.inputs.items():
            s += "\n" + _inspect_slot(inp, indent + DEFAULT_INDENT + DEFAULT_INDENT)
    if len(component.outputs) > 0:
        s += f"\n{indent}{DEFAULT_INDENT}Outputs"
        for _name, out in component.outputs.items():
            s += "\n" + _inspect_slot(out, indent + DEFAULT_INDENT + DEFAULT_INDENT)
    return s


def _inspect_adapter(adapter, indent=""):
    s = indent + adapter.name
    return s


def _inspect_slot(slot, indent=""):
    info = None
    try:
        info = slot.info
    except FinamNoDataError:
        pass

    s = indent + slot.name
    if slot.is_static:
        s += " (static)"

    s += f"\n{_inspect_info(info, indent+DEFAULT_INDENT)}"

    return s


def _inspect_info(info, indent=""):
    s = ""
    if info is None:
        s += f"{indent}?"
    else:
        s += f"\n{indent}grid: {info.grid}"
        for k, v in info.meta.items():
            s += f"\n{indent}{k}: {v}"

    return s



"""Grid abstract base classes for FINAM."""

import copy as cp
from abc import ABC, abstractmethod
from pathlib import Path

import numpy as np
from pyevtk.hl import gridToVTK, unstructuredGridToVTK

from .grid_tools import (
    CELL_DIM,
    NODE_COUNT,
    VTK_TYPE_MAP,
    CellType,
    Location,
    flatten_cells,
    gen_cells,
    gen_node_centers,
    gen_points,
    point_order,
    prepare_vtk_data,
    prepare_vtk_kwargs,
)


class GridBase(ABC):
    """Abstract grid base."""

    @property
    def name(self):
        """Grid name."""
        return self.__class__.__name__

    @property
    @abstractmethod
    def dim(self):
        """int: Dimension of the grid or data."""

    @property
    @abstractmethod
    def data_shape(self):
        """tuple: Shape of the associated data."""

    def copy(self, deep=False):
        """
        Copy of this grid.

        Parameters
        ----------
        deep : bool, optional
            If false, only a shallow copy is returned to save memory, by default False

        Returns
        -------
        Grid
            The grid copy.
        """
        return cp.deepcopy(self) if deep else cp.copy(self)

    def to_canonical(self, data):
        """Convert grid specific data to canonical form."""
        return data

    def from_canonical(self, data):
        """Convert canonical data to grid specific form."""
        return data

    # pylint: disable-next=unused-argument
    def get_transform_to(self, other):
        """Transformation between compatible grids."""
        return None

    def __repr__(self):
        return f"{self.name} ({self.dim}D) {self.data_shape}"


class Grid(GridBase):
    """Abstract grid specification."""

    valid_locations = (Location.CELLS, Location.POINTS)
    """tuple: Valid locations for the grid."""

    @property
    @abstractmethod
    def crs(self):
        """The coordinate reference system."""

    @property
    @abstractmethod
    def point_count(self):
        """int: Number of grid points."""

    @property
    @abstractmethod
    def cell_count(self):
        """int: Number of grid cells."""

    @property
    @abstractmethod
    def points(self):
        """np.ndarray: Grid points."""

    @property
    @abstractmethod
    def cells(self):
        """np.ndarray: Cell nodes as 2D array."""

    @property
    @abstractmethod
    def cell_types(self):
        """np.ndarray: Cell types."""

    @property
    def cells_connectivity(self):
        """np.ndarray: Cells connectivity in ESMF format (list of node IDs)."""
        return flatten_cells(self.cells)

    @property
    def cells_definition(self):
        """np.ndarray: Cell definition in VTK format (list of number of nodes with node IDs)."""
        return flatten_cells(
            np.squeeze(
                np.hstack(
                    (np.atleast_2d(self.cell_node_counts).T, np.atleast_2d(self.cells))
                )
            )
        )

    @property
    def cells_offset(self):
        """np.ndarray: The location of the start of each cell in cells_connectivity."""
        return np.concatenate(
            (np.array([0], dtype=int), np.cumsum(self.cell_node_counts))
        )

    @property
    def cell_centers(self):
        """np.ndarray: Grid cell centers."""
        return gen_node_centers(self)

    @property
    def cell_node_counts(self):
        """np.ndarray: Node count for each cell."""
        return NODE_COUNT[self.cell_types]

    @property
    def mesh_dim(self):
        """int: Maximal cell dimension."""
        return np.max(CELL_DIM[self.cell_types])

    @property
    @abstractmethod
    def data_location(self):
        """Location of the associated data (either CELLS or POINTS)."""

    @data_location.setter
    @abstractmethod
    def data_location(self, data_location):
        """Set location of the associated data (either CELLS or POINTS)."""

    @property
    def data_points(self):
        """Points of the associated data (either cell_centers or points)."""
        if self.data_location == Location.POINTS:
            return self.points
        return self.cell_centers

    @property
    def data_size(self):
        """int: Size of the associated data."""
        return np.prod(self.data_shape)

    @property
    @abstractmethod
    def order(self):
        """str: Data order (C-like or F-like to flatten data)."""

    @property
    @abstractmethod
    def axes_names(self):
        """list of str: Axes names (xyz order)."""

    @property
    @abstractmethod
    def axes_attributes(self):
        """list of dict: Axes attributes following the CF convention (xyz order)."""

    @property
    def data_axes_names(self):
        """list of str: Axes names of the data."""
        return ["id"]

    def compatible_with(self, other, check_location=True):
        """
        Check for compatibility with other Grid.

        Parameters
        ----------
        other : instance of Grid
            Other grid to compatibility with.
        check_location : bool, optional
            Whether to check location for equality, by default True

        Returns
        -------
        bool
            compatibility
        """
        if not isinstance(other, Grid):
            return False

        if isinstance(self, StructuredGrid) != isinstance(other, StructuredGrid):
            return False

        if not (
            self.dim == other.dim
            and self.crs == other.crs
            and self.order == other.order
            and (not check_location or self.data_location == other.data_location)
        ):
            return False

        if check_location and self.data_shape != other.data_shape:
            return False

        return (
            np.allclose(self.points, other.points)
            and np.all(self.cells == other.cells)
            and np.all(self.cell_types == other.cell_types)
        )

    def __eq__(self, other):
        return self.compatible_with(other)

    def export_vtk(
        self,
        path,
        data=None,
        cell_data=None,
        point_data=None,
        field_data=None,
        mesh_type="unstructured",
    ):
        """
        Export grid and data to a VTK file.

        Parameters
        ----------
        path : pathlike
            File path. Suffix will be replaced according to mesh type (.vtu)
        data : dict or None, optional
            Data in the corresponding shape given by name, by default None
        cell_data : dict or None, optional
            Additional cell data, by default None
        point_data : dict or None, optional
            Additional point data, by default None
        field_data : dict or None, optional
            Additional field data, by default None
        mesh_type : str, optional
            Mesh type, by default "unstructured"

        Raises
        ------
        ValueError
            If mesh type is not supported.
        """
        data = prepare_vtk_data(data, flat=True)
        kw = prepare_vtk_kwargs(
            self.data_location, data, cell_data, point_data, field_data
        )
        if mesh_type == "unstructured":
            path = str(Path(path).with_suffix(""))
            # don't create increasing axes
            points = self.points
            x = np.ascontiguousarray(points[:, 0])
            y = np.ascontiguousarray(points[:, 1] if self.dim > 1 else np.zeros_like(x))
            z = np.ascontiguousarray(points[:, 2] if self.dim > 2 else np.zeros_like(x))
            con = self.cells_connectivity
            # pyevtk only needs the ends of the cell definition
            off = self.cells_offset[1:]
            typ = VTK_TYPE_MAP[self.cell_types]
            unstructuredGridToVTK(path, x, y, z, con, off, typ, **kw)
        else:
            raise ValueError(f"export_vtk: unsupported mesh type '{mesh_type}'")


class StructuredGrid(Grid):
    """Abstract structured grid specification."""

    @property
    @abstractmethod
    def dims(self):
        """tuple: Axes lengths (xyz order)."""

    @property
    @abstractmethod
    def axes(self):
        """list of np.ndarray: Axes of the structured grid (xyz order, all increase)."""

    @property
    @abstractmethod
    def axes_reversed(self):
        """bool: Indicate reversed axes order for the associated data (zyx order)."""
        # esri grids and most netcdf files are given in inverse axes order (lat, lon)

    @property
    @abstractmethod
    def axes_increase(self):
        """list of bool: False to indicate a bottom up axis (xyz order)."""
        # esri grids and some netcdf are given bottom up (northing/lat inverted)

    @property
    @abstractmethod
    def order(self):
        """str: Point, cell and data order (C- or Fortran-like)."""
        # vtk files use Fortran-like data ordering for structured grids

    @property
    def point_count(self):
        """int: Number of grid points."""
        # allow dims entries to be 1 (flat mesh in space)
        return np.prod(self.dims)

    @property
    def cell_count(self):
        """int: Number of grid cells."""
        return np.prod(np.maximum(np.array(self.dims) - 1, 1))

    @property
    def cell_axes(self):
        """list of np.ndarray: Axes of the cell centers (xyz order, all increase)."""
        # use cell centers as stagger locations
        return [((ax[:-1] + ax[1:]) / 2) if len(ax) > 1 else ax for ax in self.axes]

    @property
    def points(self):
        """np.ndarray: Grid points in given order starting top left corner."""
        return gen_points(
            axes=self.axes,
            order=point_order(self.order, self.axes_reversed),
            axes_increase=self.axes_increase,
        )

    @property
    def cells(self):
        """np.ndarray: Cell nodes in ESMF format."""
        return gen_cells(
            dims=self.dims,
            order=point_order(self.order, self.axes_reversed),
        )

    @property
    def cell_centers(self):
        """np.ndarray: Grid cell centers in given order starting top left corner."""
        return gen_points(
            axes=self.cell_axes,
            order=point_order(self.order, self.axes_reversed),
            axes_increase=self.axes_increase,
        )

    @property
    def mesh_dim(self):
        """int: Maximal cell dimension."""
        return np.sum(np.array(self.dims) > 1)

    @property
    def cell_types(self):
        """np.ndarray: Cell types."""
        if self.mesh_dim == 0:
            return np.full(self.cell_count, CellType.VERTEX, dtype=int)
        if self.mesh_dim == 1:
            return np.full(self.cell_count, CellType.LINE, dtype=int)
        if self.mesh_dim == 2:
            return np.full(self.cell_count, CellType.QUAD, dtype=int)
        return np.full(self.cell_count, CellType.HEX, dtype=int)

    @property
    def data_axes(self):
        """list of np.ndarray: Axes as used for the data matrix."""
        axes = self.cell_axes if self.data_location == Location.CELLS else self.axes
        # reverse axes if needed
        return [
            (axes[i] if self.axes_increase[i] else axes[i][::-1])
            for i in (range(self.dim)[::-1] if self.axes_reversed else range(self.dim))
        ]

    @property
    def data_axes_names(self):
        """list of str: Axes names of the data matrix."""
        return list(
            reversed(self.axes_names) if self.axes_reversed else self.axes_names
        )

    @property
    def data_shape(self):
        """tuple: Shape of the associated data matrix."""
        dims = np.asarray(self.dims[::-1] if self.axes_reversed else self.dims)
        return tuple(
            np.maximum(dims - 1, 1) if self.data_location == Location.CELLS else dims
        )

    def compatible_with(self, other, check_location=True):
        """
        Check for compatibility with other Grid.

        Parameters
        ----------
        other : instance of Grid
            Other grid to compatibility with.
        check_location : bool, optional
            Whether to check location for equality, by default True

        Returns
        -------
        bool
            compatibility
        """
        if not isinstance(other, Grid):
            return False

        if not isinstance(other, StructuredGrid):
            return False

        if not (
            self.dim == other.dim
            and self.crs == other.crs
            and (not check_location or self.data_location == other.data_location)
        ):
            return False

        if check_location and self.data_shape != (
            other.data_shape[::-1]
            if self.axes_reversed != other.axes_reversed
            else other.data_shape
        ):
            return False

        return all(np.allclose(a, b) for a, b in zip(self.axes, other.axes))

    def __eq__(self, other):
        if not self.compatible_with(other):
            return False

        return (
            all(a == b for a, b in zip(self.axes_increase, other.axes_increase))
            and self.axes_reversed == other.axes_reversed
        )

    def export_vtk(
        self,
        path,
        data=None,
        cell_data=None,
        point_data=None,
        field_data=None,
        mesh_type="structured",
    ):
        """
        Export grid and data to a VTK file.

        Parameters
        ----------
        path : pathlike
            File path. Suffix will be replaced according to mesh type (.vtr, .vtu)
        data : dict or None, optional
            Data in the corresponding shape given by name, by default None
        cell_data : dict or None, optional
            Additional cell data, by default None
        point_data : dict or None, optional
            Additional point data, by default None
        field_data : dict or None, optional
            Additional field data, by default None
        mesh_type : str, optional
            Mesh type ("structured"/"unstructured"), by default "structured"

        Raises
        ------
        ValueError
            If mesh type is not supported.
        """
        data = prepare_vtk_data(
            data=data,
            axes_reversed=self.axes_reversed,
            axes_increase=self.axes_increase,
            flat=mesh_type == "unstructured",
            order=point_order(self.order, self.axes_reversed),
        )
        if mesh_type not in ["structured", "rectilinear"]:
            super().export_vtk(path, data, cell_data, point_data, field_data, mesh_type)
        else:
            kw = prepare_vtk_kwargs(
                self.data_location, data, cell_data, point_data, field_data
            )
            path = str(Path(path).with_suffix(""))
            x = np.ascontiguousarray(self.axes[0])
            y = np.ascontiguousarray(self.axes[1] if self.dim > 1 else np.array([0.0]))
            z = np.ascontiguousarray(self.axes[2] if self.dim > 2 else np.array([0.0]))
            gridToVTK(path, x, y, z, **kw)

    def to_canonical(self, data):
        """
        Convert grid specific data to canonical form.

        Canonical means, that data axis are in xyz order and
        following increasing axis values.

        Parameters
        ----------
        data : arraylike
            Data to convert.

        Returns
        -------
        arraylike
            Canonical Data.

        Raises
        ------
        ValueError
            When data has wrong shape.
        """
        rev = -1 if self.axes_reversed else 1
        d_shp, in_shp, shp_len = self.data_shape, np.shape(data), len(self.data_shape)
        if not np.array_equal(d_shp[::rev], in_shp[::rev][:shp_len]):
            msg = "to_canonical: data has wrong shape."
            raise ValueError(msg)
        if self.axes_reversed and np.ndim(data) > 1:
            data = np.transpose(data)
        for i, inc in enumerate(self.axes_increase):
            if not inc:
                data = np.flip(data, axis=i)
        return data

    def from_canonical(self, data):
        """
        Convert canonical data to grid specific form.

        Canonical means, that data axis are in xyz order and
        following increasing axis values.

        Parameters
        ----------
        data : arraylike
            Data to convert.

        Returns
        -------
        arraylike
            Grid specific Data.

        Raises
        ------
        ValueError
            When data has wrong shape.
        """
        rev = -1 if self.axes_reversed else 1
        d_shp, in_shp, shp_len = self.data_shape, np.shape(data), len(self.data_shape)
        if not np.array_equal(d_shp[::rev], in_shp[:shp_len]):
            msg = "from_canonical: data has wrong shape."
            raise ValueError(msg)
        for i, inc in enumerate(self.axes_increase):
            if not inc:
                data = np.flip(data, axis=i)
        if self.axes_reversed and np.ndim(data) > 1:
            data = np.transpose(data)
        return data

    def get_transform_to(self, other):
        """
        Get transformation for compatible grids.

        Parameters
        ----------
        other : instance of Grid
            Other grid to compatibility with.

        Returns
        -------
        callable
            data transformation
        """
        if not self.compatible_with(other):
            raise ValueError("get_transform_to: grids are not compatible.")

        def trans(data):
            """Transformation."""
            # could be optimized
            return other.from_canonical(self.to_canonical(data))

        # only use trans if grids are compatible but NOT equal
        return None if self == other else trans



"""
Unit tests for data info propagation.
"""

import unittest
from datetime import datetime, timedelta

import numpy as np

from finam import (
    UNITS,
    Composition,
    FinamMetaDataError,
    Info,
    Location,
    TimeComponent,
    UniformGrid,
)
from finam import data as tools
from finam.components.generators import CallbackGenerator


class MockupConsumer(TimeComponent):
    def __init__(self, time, units):
        super().__init__()
        self.time = time
        self.step = timedelta(days=1)
        self.units = units
        self.data = None

    def _next_time(self):
        return self.time + self.step

    def _initialize(self):
        self.inputs.add(name="Input", time=self.time, grid=None, units=self.units)
        self.create_connector(pull_data=["Input"])

    def _connect(self, start_time):
        self.try_connect(start_time)

        data = self.connector.in_data["Input"]
        if data is not None:
            self.data = data

    def _validate(self):
        pass

    def _update(self):
        self.data = self.inputs["Input"].pull_data(self.time)
        self.time += self.step

    def _finalize(self):
        pass


class TestUnits(unittest.TestCase):
    def test_units(self):
        time = datetime(2000, 1, 1)

        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10), spacing=(2.0, 2.0, 2.0), data_location=Location.POINTS
            ),
            units=UNITS.meter,
        )

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        source = CallbackGenerator(
            callbacks={"Output": (lambda t: in_data, in_info)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )
        sink = MockupConsumer(datetime(2000, 1, 1), UNITS.kilometer)

        composition = Composition([source, sink])

        (source.outputs["Output"] >> sink.inputs["Input"])

        composition.connect()

        self.assertEqual(sink.inputs["Input"].info.meta, {"units": UNITS.kilometer})
        self.assertEqual(tools.get_units(sink.data), UNITS.kilometer)
        self.assertEqual(tools.get_magnitude(sink.data)[0, 0, 0], 0.001)

    def test_units_fail(self):
        time = datetime(2000, 1, 1)

        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10), spacing=(2.0, 2.0, 2.0), data_location=Location.POINTS
            ),
            units=UNITS.meter,
        )

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        source = CallbackGenerator(
            callbacks={"Output": (lambda t: in_data, in_info)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        sink = MockupConsumer(datetime(2000, 1, 1), UNITS.seconds)

        composition = Composition([source, sink])

        (source.outputs["Output"] >> sink.inputs["Input"])

        with self.assertRaises(FinamMetaDataError):
            composition.run(start_time=time, end_time=datetime(2000, 1, 2))


if __name__ == "__main__":
    unittest.main()



"""
Interface definitions for the coupling framework.

Interfaces
==========

.. autosummary::
   :toctree: generated

    :noindex: IComponent
    :noindex: ITimeComponent
    :noindex: IAdapter
    :noindex: IInput
    :noindex: IOutput
    :noindex: ComponentStatus
    :noindex: Loggable
    :noindex: NoBranchAdapter
"""
import logging
from abc import ABC, abstractmethod
from enum import Enum


class ComponentStatus(Enum):
    """Status for components."""

    CREATED = 0
    INITIALIZED = 1
    CONNECTING = 2
    CONNECTING_IDLE = 3
    CONNECTED = 4
    VALIDATED = 5
    UPDATED = 6
    FINISHED = 7
    FINALIZED = 8
    FAILED = 9


class Loggable(ABC):
    """Loggable component."""

    @abstractmethod
    def __init__(self):
        self._logger = None

    @property
    @abstractmethod
    def uses_base_logger_name(self):
        """Whether this class has a ``base_logger_name`` attribute."""

    @property
    @abstractmethod
    def logger_name(self):
        """Logger name."""

    @property
    def logger(self):
        """Logger for this component."""
        if self._logger is None:
            if (
                self.uses_base_logger_name
                and hasattr(self, "base_logger_name")
                and getattr(self, "base_logger_name") is None
            ):
                return logging.getLogger(self.logger_name)
            self._logger = logging.getLogger(self.logger_name)
        return self._logger


class IComponent(ABC):
    """Interface for components."""

    @abstractmethod
    def initialize(self):
        """Initialize the component.

        After the method call, the component's inputs and outputs must be available,
        and the component should have :attr:`.status` :attr:`.ComponentStatus.INITIALIZED`.
        """

    @abstractmethod
    def connect(self, start_time):
        """Connect exchange data and metadata with linked components.

        The method can be called multiple times if there are failed pull attempts.

        After each method call, the component should have :attr:`.status` :attr:`.ComponentStatus.CONNECTED` if
        connecting was completed, :attr:`.ComponentStatus.CONNECTING` if some but not all required initial input(s)
        could be pulled, and :attr:`.ComponentStatus.CONNECTING_IDLE` if nothing could be pulled.

        Parameters
        ----------
        start_time : :class:`datetime <datetime.datetime>`
            The composition's starting time.
            Can be before the component's actual time.
        """

    @abstractmethod
    def validate(self):
        """Validate the correctness of the component's settings and coupling.

        After the method call, the component should have :attr:`.status` :attr:`.ComponentStatus.VALIDATED`.
        """

    @abstractmethod
    def update(self):
        """Update the component by one time step.
        Push new values to outputs.

        After the method call, the component should have :attr:`.status`
        :attr:`.ComponentStatus.UPDATED` or :attr:`.ComponentStatus.FINISHED`.
        """

    @abstractmethod
    def finalize(self):
        """Finalize and clean up the component.

        After the method call, the component should have :attr:`.status` :attr:`.ComponentStatus.FINALIZED`.
        """

    @property
    @abstractmethod
    def name(self):
        """Component name."""

    @property
    @abstractmethod
    def inputs(self):
        """IOList: The component's inputs."""

    @property
    @abstractmethod
    def outputs(self):
        """IOList: The component's outputs."""

    @property
    @abstractmethod
    def status(self):
        """The component's current status."""

    @property
    @abstractmethod
    def metadata(self):
        """
        The component's meta data.
        Will only be called after the connect phase of the composition.
        """


class ITimeComponent(IComponent, ABC):
    """Interface for components with a time step."""

    @property
    @abstractmethod
    def time(self):
        """The component's current simulation time."""

    @property
    @abstractmethod
    def next_time(self):
        """The component's predicted simulation time of the next pulls.

        Can be ``None`` if the component has no inputs.
        """


class IInput(ABC):
    """Interface for input slots."""

    @property
    @abstractmethod
    def name(self):
        """Input name."""

    @property
    @abstractmethod
    def is_static(self):
        """Whether the input is static"""

    @property
    @abstractmethod
    def info(self):
        """Info: The input's data info."""

    @property
    @abstractmethod
    def needs_pull(self):
        """bool: if the input needs pull."""

    @property
    @abstractmethod
    def needs_push(self):
        """bool: if the input needs push."""

    @property
    @abstractmethod
    def source(self):
        """Get the input's source output or adapter

        Returns
        -------
        :class:`.IOutput`
            The input's source.
        """

    @source.setter
    @abstractmethod
    def source(self, source):
        """Set the input's source output or adapter

        Parameters
        ----------
        source : :class:`.IOutput`
            source output or adapter
        """

    @abstractmethod
    def source_updated(self, time):
        """Informs the input that a new output is available.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the notification.
        """

    @abstractmethod
    def pull_data(self, time, target):
        """Retrieve the data from the input's source.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time to get the data for.
        target : :class:`.IInput` or None
            Requesting end point of this pull.

        Returns
        -------
        :class:`pint.Quantity`
            Data set for the given simulation time.
        """

    @abstractmethod
    def ping(self):
        """Pings upstream to inform outputs about the number of connected inputs.

        Must be called after linking and before the connect phase.
        """

    @abstractmethod
    def exchange_info(self, info):
        """Exchange the data info with the input's source.

        Parameters
        ----------
        info : :class:`.Info`
            request parameters

        Returns
        -------
        dict
            delivered parameters
        """


class IOutput(ABC):
    """Interface for output slots."""

    @property
    @abstractmethod
    def name(self):
        """Output name."""

    @property
    @abstractmethod
    def time(self):
        """The output's time of the latest available data"""

    @property
    @abstractmethod
    def is_static(self):
        """Whether the output is static"""

    @property
    @abstractmethod
    def info(self):
        """Info: The output's data info.

        Raises
        ------
        FinamNoDataError
            Raises the error if infos were not yet exchanged
        """

    @property
    @abstractmethod
    def needs_pull(self):
        """bool: if the output needs pull."""

    @property
    @abstractmethod
    def needs_push(self):
        """bool: if the output needs push."""

    @property
    @abstractmethod
    def memory_limit(self):
        """The memory limit for this slot"""

    @memory_limit.setter
    @abstractmethod
    def memory_limit(self, limit):
        """The memory limit for this slot"""

    @property
    @abstractmethod
    def memory_location(self):
        """The memory-mapping location for this slot"""

    @memory_location.setter
    @abstractmethod
    def memory_location(self, directory):
        """The memory-mapping location for this slot"""

    @abstractmethod
    def has_info(self):
        """Returns if the output has a data info.

        The info is not required to be validly exchanged.
        """

    @abstractmethod
    def add_target(self, target):
        """Add a target input or adapter for this output.

        Parameters
        ----------
        target : :class:`.IInput`
            The target to add.
        """

    @property
    @abstractmethod
    def targets(self):
        """Get target inputs and adapters for this output.

        Returns
        -------
        list
            List of targets.
        """

    @abstractmethod
    def pinged(self, source):
        """Called when receiving a ping from a downstream input.

        Parameters
        ----------

        source : :class:`.IInput`
            Pinging target end point
        """

    @abstractmethod
    def push_data(self, data, time):
        """Push data into the output.

        Should notify targets, and can handle the provided date.

        Parameters
        ----------
        data : array_like
            Data set to push.
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the data set.
        """

    @abstractmethod
    def push_info(self, info):
        """Push data info into the output.

        Parameters
        ----------
        info : :class:`.Info`
            Delivered data info
        """

    @abstractmethod
    def notify_targets(self, time):
        """Notify all targets by calling their ``source_updated(time)`` method.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the simulation.
        """

    @abstractmethod
    def get_data(self, time, target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time to get the data for.
        target : :class:`.IInput`
            Requesting end point of this pull

        Returns
        -------
        :class:`pint.Quantity`
            data-set for the requested time.

        Raises
        ------
        FinamNoDataError
            Raises the error if no data is available
        """

    @abstractmethod
    def get_info(self, info):
        """Exchange and get the output's data info.

        Parameters
        ----------
        info : :class:`.Info`
            Requested data info

        Returns
        -------
        dict
            Delivered data info

        Raises
        ------
        FinamNoDataError
            Raises the error if no info is available
        """

    @abstractmethod
    def chain(self, other):
        """Chain outputs and adapters.

        Parameters
        ----------
        other : :class:`.IInput` or :class:`.IAdapter`
            The adapter or input to add as target to this output.

        Returns
        -------
        :class:`.IOutput`
            The last element of the chain.
        """

    @abstractmethod
    def finalize(self):
        """Finalize the output"""

    def __rshift__(self, other):
        return self.chain(other)


class IAdapter(IInput, IOutput, ABC):
    """Interface for adapters."""

    @property
    @abstractmethod
    def metadata(self):
        """
        The adapter's meta data.
        Will only be called after the connect phase of the composition.
        """


class NoBranchAdapter:
    """Interface to mark adapters as allowing only a single end point."""


class NoDependencyAdapter:
    """Interface to mark adapters as breaking time dependencies between components."""


class ITimeDelayAdapter(ABC):
    """Interface for adapters that manipulate the request time."""

    @abstractmethod
    def with_delay(self, time):
        """Get the manipulated time for a given request time.

        Parameters
        ----------

        time : :class:`datetime <datetime.datetime>`
            The original request time.

        Returns
        -------

        :class:`datetime <datetime.datetime>`
            The time as manipulated by the adapter
        """



"""
Basic data transformation adapters.
"""

import numpy as np

from ..data.grid_spec import NoGrid
from ..data.tools import Mask, get_magnitude, mask_specified
from ..errors import FinamMetaDataError
from ..sdk import Adapter
from ..tools.log_helper import ErrorLogger

__all__ = [
    "Callback",
    "Scale",
    "ValueToGrid",
    "GridToValue",
]


class Callback(Adapter):
    """Transform data using a callback.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.Callback(
            callback=lambda data, t: data * 2,
        )

    Parameters
    ----------
    callback : callable
        A callback ``callback(data, time)``, returning the transformed data.
    """

    def __init__(self, callback):
        super().__init__()
        self.callback = callback

    def _get_data(self, time, target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : datetime
            Simulation time to get the data for.

        Returns
        -------
        array_like
            data-set for the requested time.
        """
        d = self.pull_data(time, target)
        return self.callback(d, time)


class Scale(Adapter):
    """Scales the input.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.Scale(scale=0.5)

    Parameters
    ----------
    scale : float
        Scale factor.
    """

    def __init__(self, scale):
        super().__init__()
        self.scale = scale

    def _get_data(self, time, target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : datetime
            simulation time to get the data for.

        Returns
        -------
        array_like
            data-set for the requested time.
        """
        d = self.pull_data(time, target)
        return d * self.scale


class ValueToGrid(Adapter):
    """Convert a scalar value to a Matrix filled with that value.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.ValueToGrid(
            grid=fm.UniformGrid(dims=(10, 20))
        )

        adapter = fm.adapters.ValueToGrid(grid=None)

    Parameters
    ----------
    grid: :any:`Grid`, optional
        Grid specification to create grid for.
        Will be ``None`` by default to get it from the target.
    mask : :any:`Mask` value or valid boolean mask for :any:`MaskedArray`, optional
        masking specification of the data. Options:
            * :any:`Mask.FLEX`: data can be masked or unmasked
            * :any:`Mask.NONE`: data is unmasked and given as plain numpy array
            * valid boolean mask for MaskedArray
        Will be ``None`` by default to get it from the target.
    """

    def __init__(self, grid=None, mask=None):
        super().__init__()
        self.grid = grid
        self.mask = mask

    def _get_data(self, time, target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : datetime
            simulation time to get the data for.

        Returns
        -------
        array_like
            data-set for the requested time.
        """
        value = self.pull_data(time, target)
        data = np.full(self.info.grid_shape, get_magnitude(value))
        if mask_specified(self.info.mask):
            return np.ma.array(data, mask=self.info.mask)
        return data

    def _get_info(self, info):
        up_info = info.copy_with(grid=NoGrid(), mask=Mask.FLEX)
        in_info = self.exchange_info(up_info)
        out_info = in_info.copy_with(
            grid=self.grid or info.grid, mask=self.mask or info.mask, use_none=False
        )

        if info.grid is not None and info.grid != out_info.grid:
            with ErrorLogger(self.logger):
                raise FinamMetaDataError(
                    f"Grid specifications don't match. Target has {info.grid}, expected {out_info.grid}"
                )

        return out_info


class GridToValue(Adapter):
    """Convert a matrix to a scalar value using an aggregation function, e.g. ``numpy.mean``.

    Examples
    --------

    .. testcode:: constructor

        import numpy as np
        import finam as fm

        adapter = fm.adapters.GridToValue(func=np.mean)

    Parameters
    ----------
    func : callable
        A function ``func(data)``, returning the transformed data.
    """

    def __init__(self, func):
        super().__init__()
        self.func = func

    def _get_data(self, time, target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : datetime
            simulation time to get the data for.

        Returns
        -------
        array_like
            data-set for the requested time.
        """
        grid = self.pull_data(time, target)
        return self.func(get_magnitude(grid))

    def _get_info(self, info):
        info = info.copy_with(grid=None)
        in_info = self.exchange_info(info)
        out_info = in_info.copy_with(grid=NoGrid(), mask=Mask.NONE)
        return out_info



import unittest
from datetime import datetime, timedelta

import numpy as np

import finam as fm
from finam import Composition, Info, NoGrid
from finam.components import CallbackComponent, CallbackGenerator, DebugConsumer


def transform(inputs, _time):
    return {"Out1": inputs["In1"][0, ...] * 2.0}


def consume(_inputs, _time):
    return {}


class TestCallback(unittest.TestCase):
    def test_callback(self):
        start = datetime(2000, 1, 1)

        source = CallbackGenerator(
            callbacks={
                "Out1": (
                    lambda t: 1.0 + np.random.random(1)[0],
                    Info(None, grid=NoGrid()),
                )
            },
            start=start,
            step=timedelta(days=7),
        )

        trans = CallbackComponent(
            inputs={"In1": Info(None, grid=NoGrid())},
            outputs={"Out1": Info(None, grid=NoGrid())},
            callback=transform,
            start=start,
            step=timedelta(days=7),
        )

        consumer = DebugConsumer(
            inputs={"In1": Info(None, grid=NoGrid())},
            start=start,
            step=timedelta(days=7),
            log_data="DEBUG",
        )

        composition = Composition([source, trans, consumer])

        _ = source.outputs["Out1"] >> trans.inputs["In1"]
        _ = trans.outputs["Out1"] >> consumer.inputs["In1"]

        composition.connect(start)

        out_data = consumer.data["In1"]
        self.assertGreaterEqual(out_data, 2.0)
        self.assertLessEqual(out_data, 4.0)

        composition.run(end_time=datetime(2000, 3, 1))

        out_data = consumer.data["In1"]
        self.assertGreaterEqual(out_data, 2.0)
        self.assertLessEqual(out_data, 4.0)


if __name__ == "__main__":
    unittest.main()



"""
The FINAM model coupling framework.

See also these FINAM book chapters:

* :doc:`/finam-book/usage/coupling_scripts` for writing FINAM coupling scripts.
* :doc:`/finam-book/development/components` for writing FINAM components or model wrappers.
* :doc:`/finam-book/development/adapters` for writing FINAM adapters.

.. toctree::
   :hidden:

   self

Schedule
========

Driver/scheduler for creating and executing coupled model compositions.

.. autosummary::
   :toctree: generated
   :caption: Schedule

    Composition

Software development kit
========================

Implementations of FINAM interfaces for component and adapter development.

.. autosummary::
   :toctree: generated
   :caption: Software development kit

    Adapter
    CallbackInput
    CallbackOutput
    Component
    Input
    Output
    TimeComponent
    TimeDelayAdapter

Grids
=====

Grid specifications for the exchange of spatial data in FINAM.

.. autosummary::
   :toctree: generated
   :caption: Grids

    EsriGrid
    NoGrid
    RectilinearGrid
    UniformGrid
    UnstructuredGrid
    UnstructuredPoints

Grid tools
==========

Utilities for grid specifications.

.. autosummary::
   :toctree: generated
   :caption: Grid tools

    CellType
    Location

Data tools
==========

Utilities for data and metadata handling.

.. autosummary::
   :toctree: generated
   :caption: Data tools

    Info
    Mask
    UNITS

Interfaces
==========

Basic interfaces of the FINAM framework.

.. autosummary::
   :toctree: generated
   :caption: Interfaces

    ComponentStatus
    IComponent
    ITimeComponent
    IAdapter
    IInput
    IOutput
    Loggable
    NoBranchAdapter
    NoDependencyAdapter
    ITimeDelayAdapter

Errors
======

FINAM-specific error types.

.. autosummary::
   :toctree: generated
   :caption: Errors

    FinamCircularCouplingError
    FinamConnectError
    FinamDataError
    FinamLogError
    FinamMetaDataError
    FinamNoDataError
    FinamStaticDataError
    FinamStatusError
    FinamTimeError

Subpackages
===========

Built-in components, adapters and tool functions.

.. autosummary::
   :toctree: generated
   :caption: Subpackages

    adapters
    data
    components
    tools

"""
from . import adapters, components, data, interfaces, schedule, sdk, tools
from .data.grid_spec import (
    EsriGrid,
    NoGrid,
    RectilinearGrid,
    UniformGrid,
    UnstructuredGrid,
    UnstructuredPoints,
)
from .data.grid_tools import CellType, Location
from .data.tools import UNITS, Info, Mask
from .errors import (
    FinamCircularCouplingError,
    FinamConnectError,
    FinamDataError,
    FinamLogError,
    FinamMetaDataError,
    FinamNoDataError,
    FinamStaticDataError,
    FinamStatusError,
    FinamTimeError,
)
from .interfaces import (
    ComponentStatus,
    IAdapter,
    IComponent,
    IInput,
    IOutput,
    ITimeComponent,
    ITimeDelayAdapter,
    Loggable,
    NoBranchAdapter,
    NoDependencyAdapter,
)
from .schedule import Composition
from .sdk import (
    Adapter,
    CallbackInput,
    CallbackOutput,
    Component,
    Input,
    Output,
    TimeComponent,
    TimeDelayAdapter,
)

try:
    from ._version import __version__
except ModuleNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = "0.0.0.dev0"

tools.log_helper.add_logging_level("TRACE", 5)
tools.log_helper.add_logging_level("PROFILE", 15)


__all__ = ["__version__"]
__all__ += ["adapters", "components", "data", "interfaces", "schedule", "sdk", "tools"]
__all__ += [
    "IComponent",
    "ITimeComponent",
    "IAdapter",
    "IInput",
    "IOutput",
    "ComponentStatus",
    "Loggable",
    "NoBranchAdapter",
    "NoDependencyAdapter",
    "ITimeDelayAdapter",
]
__all__ += ["Composition"]
__all__ += [
    "Adapter",
    "Component",
    "TimeComponent",
    "TimeDelayAdapter",
    "CallbackInput",
    "CallbackOutput",
    "Input",
    "Output",
]
__all__ += [
    "EsriGrid",
    "NoGrid",
    "RectilinearGrid",
    "UniformGrid",
    "UnstructuredGrid",
    "UnstructuredPoints",
]
__all__ += ["CellType", "Location"]
__all__ += ["UNITS", "Info", "Mask"]
__all__ += [
    "FinamCircularCouplingError",
    "FinamConnectError",
    "FinamDataError",
    "FinamLogError",
    "FinamMetaDataError",
    "FinamNoDataError",
    "FinamStaticDataError",
    "FinamStatusError",
    "FinamTimeError",
]



import unittest

import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

from finam import CellType, EsriGrid, NoGrid, UniformGrid, UnstructuredGrid
from finam.data.grid_tools import (
    CELL_DIM,
    INV_VTK_TYPE_MAP,
    VTK_CELL_DIM,
    VTK_TYPE_MAP,
    check_axes_monotonicity,
    check_axes_uniformity,
    flatten_cells,
    gen_axes,
    gen_cells,
    gen_node_centers,
    gen_points,
    get_cells_matrix,
    order_map,
    point_order,
)


class TestGridTools(unittest.TestCase):
    def test_point_order(self):
        self.assertEqual(point_order(order="C", axes_reversed=False), "C")
        self.assertEqual(point_order(order="C", axes_reversed=True), "F")
        self.assertEqual(point_order(order="F", axes_reversed=False), "F")
        self.assertEqual(point_order(order="F", axes_reversed=True), "C")

    def test_order_map(self):
        assert_array_equal(order_map((2, 3), of="C", to="C"), [0, 1, 2, 3, 4, 5])
        assert_array_equal(order_map((2, 3), of="F", to="F"), [0, 1, 2, 3, 4, 5])
        assert_array_equal(order_map((2, 3), of="F", to="C"), [0, 2, 4, 1, 3, 5])
        assert_array_equal(order_map((2, 3), of="C", to="F"), [0, 3, 1, 4, 2, 5])

    def test_gen_node_centers(self):
        uniform = UniformGrid((2, 3), (2.0, 2.0, 2.0), (1.0, 1.0, 1.0))
        assert_allclose(gen_node_centers(uniform), [[2.0, 2.0], [2.0, 4.0]])

        unstruct = UnstructuredGrid(
            points=[[0.0, 0.0], [0.0, 3.0], [3.0, 0.0]],
            cells=[0, 1, 2],
            cell_types=[CellType.TRI.value],
        )
        assert_allclose(gen_node_centers(unstruct), [[1.0, 1.0]])

        unstruct = UnstructuredGrid(
            points=[[0.0, 0.0], [0.0, 2.0], [2.0, 2.0], [2.0, 0.0]],
            cells=[0, 1, 2, 3],
            cell_types=[CellType.QUAD.value],
        )
        assert_allclose(gen_node_centers(unstruct), [[1.0, 1.0]])

    def test_gen_axes(self):
        axes = gen_axes((3, 4), (2.0, 3.0), (1.0, 1.0))
        assert_allclose(axes[0], [1.0, 3.0, 5.0])
        assert_allclose(axes[1], [1.0, 4.0, 7.0, 10.0])

        axes = gen_axes((3, 4), (2.0, 3.0), (1.0, 1.0), (False, False))
        assert_allclose(axes[0], [5.0, 3.0, 1.0])
        assert_allclose(axes[1], [10.0, 7.0, 4.0, 1.0])

    def test_gen_points(self):
        axes = gen_axes((3, 4), (2.0, 3.0), (1.0, 1.0))
        points = gen_points(axes)
        assert_allclose(
            points[:6],
            [[1.0, 1.0], [3.0, 1.0], [5.0, 1.0], [1.0, 4.0], [3.0, 4.0], [5.0, 4.0]],
        )

        axes = gen_axes((3, 4), (2.0, 3.0), (1.0, 1.0))
        points = gen_points(axes, order="C")
        assert_allclose(
            points[:6],
            [[1.0, 1.0], [1.0, 4.0], [1.0, 7.0], [1.0, 10.0], [3.0, 1.0], [3.0, 4.0]],
        )

        axes = gen_axes((3, 4), (2.0, 3.0), (1.0, 1.0), (False, False))
        points = gen_points(axes, order="C", axes_increase=(False, False))
        assert_allclose(
            points[:6],
            [[1.0, 1.0], [1.0, 4.0], [1.0, 7.0], [1.0, 10.0], [3.0, 1.0], [3.0, 4.0]],
        )

    def test_gen_cells(self):
        assert_array_equal(gen_cells((), order="F"), [[0]])
        assert_array_equal(gen_cells((1,), order="F"), [[0]])
        assert_array_equal(gen_cells((2,), order="F"), [[0, 1]])

        assert_array_equal(gen_cells((2, 2, 2), order="F"), [[6, 7, 5, 4, 2, 3, 1, 0]])

        # 0---1---2
        # |   |   |
        # 3-->4-->5
        assert_array_equal(gen_cells((3, 2), order="F"), [[3, 4, 1, 0], [4, 5, 2, 1]])
        # 0---1
        # |   |
        # 2-->3
        # |   |
        # 4-->5
        assert_array_equal(gen_cells((2, 3), order="F"), [[2, 3, 1, 0], [4, 5, 3, 2]])
        # 0---2---4
        # |   |   |
        # 1-->3-->5
        assert_array_equal(gen_cells((3, 2), order="C"), [[1, 3, 2, 0], [3, 5, 4, 2]])
        # 0---3
        # |   |
        # 1-->4
        # |   |
        # 2-->5
        assert_array_equal(gen_cells((2, 3), order="C"), [[1, 4, 3, 0], [2, 5, 4, 1]])

    def test_check_axes_monotonicity(self):
        axes = [np.asarray([0, 1, 2]), np.asarray([0, 1, 2, 3])]
        assert_array_equal(check_axes_monotonicity(axes), [True, True])
        assert_array_equal(axes[0], [0, 1, 2])
        assert_array_equal(axes[1], [0, 1, 2, 3])

        axes = [np.asarray([2, 1, 0]), np.asarray([0, 1, 2, 3])]
        assert_array_equal(check_axes_monotonicity(axes), [False, True])
        assert_array_equal(axes[0], [0, 1, 2])
        assert_array_equal(axes[1], [0, 1, 2, 3])

        axes = [np.asarray([0, 1, 2]), np.asarray([3, 2, 1, 0])]
        assert_array_equal(check_axes_monotonicity(axes), [True, False])
        assert_array_equal(axes[0], [0, 1, 2])
        assert_array_equal(axes[1], [0, 1, 2, 3])

        axes = [np.asarray([2, 1, 0]), np.asarray([3, 2, 1, 0])]
        assert_array_equal(check_axes_monotonicity(axes), [False, False])
        assert_array_equal(axes[0], [0, 1, 2])
        assert_array_equal(axes[1], [0, 1, 2, 3])

        axes = [np.asarray([2, 0, 1]), np.asarray([0, 1, 2, 3])]
        with self.assertRaises(ValueError):
            check_axes_monotonicity(axes)

        axes = [np.asarray([0, 1, 2]), np.asarray([0, 2, 1, 3])]
        with self.assertRaises(ValueError):
            check_axes_monotonicity(axes)

    def test_errors(self):
        # wrong len(axes_increase)
        with self.assertRaises(ValueError):
            UniformGrid((4, 2), axes_increase=[False])
        # wrong mesh_type
        with self.assertRaises(ValueError):
            grid = UniformGrid((2, 2))
            grid.export_vtk("test", mesh_type="wrong")
        # wrong mesh_type
        with self.assertRaises(ValueError):
            grid = UniformGrid((2, 2), data_location=None)
            grid.data_points

    def test_flatten_cells(self):
        assert_array_equal(flatten_cells(np.asarray([0, 1, 2, 3])), [0, 1, 2, 3])
        assert_array_equal(
            flatten_cells(np.asarray([[0, 1, 2, 3], [4, 5, 6, 7]])),
            [0, 1, 2, 3, 4, 5, 6, 7],
        )

    def test_axes_uniformity(self):
        assert_array_equal(
            check_axes_uniformity([np.asarray([0, 1, 2, 3]), np.asarray([0, 2, 3])]),
            [1, float("nan")],
        )

    def test_compatibility(self):
        grid1 = EsriGrid(nrows=2, ncols=3)
        grid2 = UniformGrid((4, 3))
        data1 = np.arange(6).reshape((2, 3))

        self.assertTrue(grid1.compatible_with(grid2))

        trans = grid1.get_transform_to(grid2)
        trans2 = grid2.get_transform_to(grid1)
        self.assertIsNotNone(trans)
        data2 = trans(data1)
        data3 = trans2(data2)
        assert_array_equal(data1.shape, data2.T.shape)
        assert_array_equal(data1[0], data2.T[1])
        assert_array_equal(data1[1], data2.T[0])
        assert_array_equal(data1, data3)

        grid1 = EsriGrid(nrows=2, ncols=3, cellsize=2.0)
        self.assertFalse(grid1.compatible_with(grid2))

        grid1 = EsriGrid(nrows=3, ncols=3)
        self.assertFalse(grid1.compatible_with(grid2))

        grid1 = EsriGrid(nrows=2, ncols=3)
        # layer height info results in 3D grid
        grid2 = UniformGrid((4, 3, 1))
        self.assertFalse(grid1.compatible_with(grid2))

    def test_cells_definition(self):
        # number of nodes with node IDs
        c_def = np.array(
            [3, 0, 6, 5]
            + [3, 0, 1, 6]
            + [4, 1, 2, 7, 6]
            + [3, 2, 8, 7]
            + [3, 2, 3, 8]
            + [4, 3, 4, 9, 8]
        )
        # node IDs
        c_con = np.array(
            [0, 6, 5] + [0, 1, 6] + [1, 2, 7, 6] + [2, 8, 7] + [2, 3, 8] + [3, 4, 9, 8]
        )
        # cell offsets: start of cell definition in connectivity plus size of connectivity
        c_off = np.array([0, 3, 6, 10, 13, 16, 20])
        # node IDs as matrix
        c_mat = np.array(
            [
                [0, 6, 5, -1],
                [0, 1, 6, -1],
                [1, 2, 7, 6],
                [2, 8, 7, -1],
                [2, 3, 8, -1],
                [3, 4, 9, 8],
            ]
        )
        cell_types = np.array([2, 2, 3, 2, 2, 3])
        pnts = np.array(
            [
                [0.0, 0.0],
                [1.0, 0.0],
                [2.0, 0.0],
                [3.0, 0.0],
                [4.0, 0.0],
                [0.0, 1.0],
                [1.0, 1.0],
                [2.0, 1.0],
                [3.0, 1.0],
                [4.0, 1.0],
            ]
        )
        grid = UnstructuredGrid(points=pnts, cells=c_mat, cell_types=cell_types)

        assert_array_equal(grid.cells_connectivity, c_con)
        assert_array_equal(grid.cells_definition, c_def)
        assert_array_equal(grid.cells_offset, c_off)
        self.assertEqual(grid.cells_offset[0], 0)
        self.assertEqual(grid.cells_offset[-1], len(c_con))
        assert_array_equal(get_cells_matrix(grid.cell_types, c_def), c_mat)
        assert_array_equal(
            get_cells_matrix(grid.cell_types, c_con, connectivity=True), c_mat
        )

    def test_vtk_maps(self):
        test_dims = VTK_CELL_DIM[VTK_TYPE_MAP]
        assert_array_equal(test_dims, CELL_DIM)

        test_type = INV_VTK_TYPE_MAP[VTK_TYPE_MAP]
        assert_array_equal(test_type, range(6))


if __name__ == "__main__":
    unittest.main()



"""Grid tools for FINAM."""

from enum import Enum, IntEnum
from math import isclose, nan

import numpy as np


def point_order(order, axes_reversed=False):
    """
    Determine apparent point order incorporating axes order reversion.

    Parameters
    ----------
    order : str
        Point and cell ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "F"
    axes_reversed : arraylike or None, optional
        Indicate reversed axes order for the associated data, by default False

    Returns
    -------
    str
        Apparent point ordering.
    """
    reverse_order = {"C": "F", "F": "C"}
    return reverse_order[order] if axes_reversed else order


def order_map(shape, of="F", to="C"):
    """
    Generate order mapping.

    Parameters
    ----------
    shape : tuple
        Array shape of interest.
    of : str, optional
        Starting ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "F"
    to : str, optional
        Target ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "C"

    Returns
    -------
    np.ndarray
        Mapping indices.
    """
    size = np.prod(shape)
    return np.arange(size, dtype=int).reshape(shape, order=of).reshape(-1, order=to)


def gen_node_centers(grid):
    """
    Calculate the node centers of the given grid cells.

    Parameters
    ----------
    grid : Grid
        Grid to take the cells from.

    Returns
    -------
    np.ndarray
        Centroids for all cells.
    """
    unique_cell_types = np.unique(grid.cell_types)
    result = np.empty((grid.cell_count, grid.dim), dtype=float)
    for ctype in unique_cell_types:
        # select current cell type
        sel = grid.cell_types == ctype
        points = grid.points[grid.cells[sel][:, : NODE_COUNT[ctype]]]
        result[sel] = np.mean(points, axis=1)
    return result


def gen_axes(dims, spacing, origin, axes_increase=None):
    """
    Generate uniform axes.

    Parameters
    ----------
    dims : iterable
        Dimensions of the uniform grid for each direction.
    spacing : iterable
        Spacing of the uniform in each dimension. Must be positive.
    origin : iterable
        Origin of the uniform grid.
    axes_increase : arraylike or None, optional
        False to indicate a bottom up axis (in xyz order), by default None

    Returns
    -------
    list of np.ndarray
        Axes of the uniform grid.
    """
    if axes_increase is None:
        axes_increase = np.full(len(dims), True, dtype=bool)
    if len(axes_increase) != len(dims):
        raise ValueError("gen_axes: wrong length of 'axes_increase'")
    axes = []
    for i, d in enumerate(dims):
        axes.append(np.arange(d) * spacing[i] + origin[i])
        if not axes_increase[i]:
            axes[i] = axes[i][::-1]
    return axes


def gen_points(axes, order="F", axes_increase=None):
    """
    Generate points from axes of a rectilinear grid.

    Parameters
    ----------
    axes : list of np.ndarray
        Axes defining the coordinates in each direction (xyz order).
    order : str, optional
        Point and cell ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "F"
    axes_increase : arraylike or None, optional
        False to indicate a bottom up axis (in xyz order), by default None

    Returns
    -------
    np.ndarray
        Points of the grid in given order and orientation.
    """
    if axes_increase is None:
        axes_increase = np.full(len(axes), True, dtype=bool)
    axes = list(axes)
    for i, inc in enumerate(axes_increase):
        if not inc:
            axes[i] = axes[i][::-1]
    dim = len(axes)
    # append empty dimensions
    for _ in range(dim, 3):
        axes.append(np.zeros(1, dtype=float))
    x_dim = len(axes[0])
    y_dim = len(axes[1])
    z_dim = len(axes[2])
    pnt_cnt = x_dim * y_dim * z_dim
    x_id, y_id, z_id = np.mgrid[0:x_dim, 0:y_dim, 0:z_dim]
    points = np.empty((pnt_cnt, 3), dtype=float)
    # VTK sorts points and cells in Fortran order
    points[:, 0] = axes[0][x_id.reshape(-1, order=order)]
    points[:, 1] = axes[1][y_id.reshape(-1, order=order)]
    points[:, 2] = axes[2][z_id.reshape(-1, order=order)]
    return points[:, :dim]


def gen_cells(dims, order="F"):
    """
    Generate cells from dimensions of a structured grid.

    Parameters
    ----------
    dims : iterable
        Dimensions of the structured grid for each direction.
    order : str, optional
        Point and cell ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "F"

    Returns
    -------
    np.ndarray
        Cell definitions containing the list of node IDs for each cell.
    """
    # sort out empty dimensions
    c_dim = [d - 1 for d in dims if d > 1]
    c_cnt = int(np.prod(c_dim))
    mesh_dim = len(c_dim)
    c_rng = np.arange(c_cnt, dtype=int)
    if mesh_dim == 0:
        # cells are vertices in 0D
        c = np.array([[0]], dtype=int)
    elif mesh_dim == 1:
        # cells are lines in 1D
        c = np.empty((c_cnt, 2), dtype=int)
        c[:, 0] = c_rng
        c[:, 1] = c[:, 0] + 1
    elif mesh_dim == 2:
        # cells are quad in 2D
        c = np.empty((c_cnt, 4), dtype=int)
        # top left corner (last node in cell definition)
        c[:, 3] = c_rng
        c[:, 3] += c_rng // c_dim[0]
        # top right corner one ID higher than left
        c[:, 2] = c[:, 3] + 1
        # lower right corner two IDs higher than top left in next row
        c[:, 1] = c[:, 3] + 2 + c_dim[0]
        # lower left corner one ID lower than right
        c[:, 0] = c[:, 1] - 1
    else:
        # cells are hex in 3D
        c = np.empty((c_cnt, 8), dtype=int)
        # ? should upper and lower layer be swapped?
        # upper layer
        c[:, 7] = c_rng
        c[:, 7] += (c_dim[0] + c_dim[1] + 1) * (c_rng // (c_dim[0] * c_dim[1]))
        c[:, 7] += (c_rng % (c_dim[0] * c_dim[1])) // c_dim[0]
        c[:, 6] = c[:, 7] + 1
        c[:, 5] = c[:, 7] + 2 + c_dim[0]
        c[:, 4] = c[:, 5] - 1
        # lower layer
        c[:, 3] = c[:, 7] + (1 + c_dim[0]) * (1 + c_dim[1])
        c[:, 2] = c[:, 3] + 1
        c[:, 1] = c[:, 3] + 2 + c_dim[0]
        c[:, 0] = c[:, 1] - 1
    if order == "C" and mesh_dim > 1:
        # inverse reorder point ids
        c = order_map(dims, of="C", to="F")[c]
        # reorder cells
        c = c[order_map(c_dim, of="F", to="C")]
    return c


def check_axes_monotonicity(axes):
    """
    Check axes to be strictly monotonic, and makes them strictly monotonic increasing.

    Parameters
    ----------
    axes : list of np.ndarray
        Axes defining the coordinates in each direction (xyz order).
        Will be modified inplace to be increasing.

    Returns
    -------
    axes_increase : list of bool
        False to indicate a bottom up axis.

    Raises
    ------
    ValueError
        In case an axis in not strictly monotonic.
    """
    axes_increase = np.empty(len(axes), dtype=bool)
    for i, ax in enumerate(axes):
        if len(ax) == 1:
            axes_increase[i] = True
            continue
        dx = ax[1:] - ax[:-1]
        if np.all(dx > 0):
            axes_increase[i] = True
        elif np.all(dx < 0):
            ax[:] = ax[::-1]
            axes_increase[i] = False
        else:
            raise ValueError(f"Grid: axes[{i}] not strictly monotonic.")
    return axes_increase


def check_axes_uniformity(axes):
    """
    Check axes to be uniform.

    Parameters
    ----------
    axes : list of np.ndarray
        Axes defining the coordinates in each direction (xyz order).

    Returns
    -------
    is_uniform : list of float
        Spacing or NaN for each axis. NaN indicates non-uniformity
    """
    return [check_uniformity(ax) for ax in axes]


def check_uniformity(values):
    """Checks for uniform spacing of values

    Parameters
    ----------
    values : np.ndarray
        Values to check.

    Returns
    -------
    is_uniform : float
        Average spacing, of NaN if not uniform.
    """
    delta = None
    for i in range(len(values) - 1):
        d = values[i + 1] - values[i]
        if delta is None:
            delta = d
        elif not isclose(delta, d):
            return nan

    return (values[-1] - values[0]) / (len(values) - 1)


def prepare_vtk_kwargs(data_location, data, cell_data, point_data, field_data):
    """
    Prepare keyword arguments for evtk routines.

    Parameters
    ----------
    data_location : Location
        Data location in the grid, by default Location.CELLS
    data : dict or None
        Data in the corresponding shape given by name
    cell_data : dict or None
        Additional cell data
    point_data : dict or None
        Additional point data
    field_data : dict or None
        Additional field data

    Returns
    -------
    dict
        Keyword arguments.
    """
    cdat = data_location == Location.CELLS
    kw = {"cellData": None, "pointData": None, "fieldData": None}
    kw["cellData" if cdat else "pointData"] = data
    if kw["cellData"]:
        kw["cellData"].update(cell_data if cell_data is not None else {})
    else:
        kw["cellData"] = cell_data
    if kw["pointData"]:
        kw["pointData"].update(point_data if point_data is not None else {})
    else:
        kw["pointData"] = point_data
    kw["fieldData"] = field_data
    return kw


def prepare_vtk_data(
    data, axes_reversed=False, axes_increase=None, flat=False, order="F"
):
    """
    Prepare data dictionary for VTK export.

    Parameters
    ----------
    data : dict or None
        Dictionary containing data arrays by name.
    axes_reversed : bool, optional
        Indicate reversed axes order for the associated data, by default False
    axes_increase : arraylike or None, optional
        False to indicate a bottom up axis (xyz order), by default None
    flat : bool, optional
        True to flatten data, by default False
    order : str, optional
        Point and cell ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "F"

    Returns
    -------
    dict or None
        Prepared data.
    """
    if data is not None:
        data = dict(data)
        for name, value in data.items():
            data[name] = np.ascontiguousarray(
                _prepare(value, axes_reversed, axes_increase, flat, order)
            )
    return data


def _prepare(data, axes_reversed, axes_increase, flat, order):
    if axes_increase is not None and data.ndim != len(axes_increase):
        raise ValueError("prepare_vtk_data: data has wrong dimension.")
    if axes_increase is None:
        axes_increase = np.full(data.ndim, True, dtype=bool)
    if axes_reversed and data.ndim > 1:
        data = data.T
    for i, inc in enumerate(axes_increase):
        # only flip if not converting to unstructured
        if not (inc or flat):
            data = np.flip(data, axis=i)
    # get 3D or flat shape
    shape = -1 if flat else (data.shape + (1,) * (3 - data.ndim))
    return data.reshape(shape, order=order)


def flatten_cells(cells):
    """
    Flatten cells array.

    Parameters
    ----------
    cells : np.ndarray
        Cells given as 2D array containing cell defining node IDs.
        -1 will be interpreted as used entries.

    Returns
    -------
    np.ndarray
        All cell definitions concatenated.
    """
    if cells.ndim == 1:
        return cells
    data = cells.ravel()
    # unused entries in "cells" marked with "-1"
    return data.compress(data != -1)


def get_cells_matrix(cell_types, cells, connectivity=False):
    """
    Create the cells matrix as used in the Grid class.

    Parameters
    ----------
    cell_types : np.ndarray
        Cell types.

    cells : np.ndarray
        Either cell definitions given as list of number of nodes with node IDs:
        ``[n0, p0_0, p0_1, ..., p0_n, n1, p1_0, p1_1, ..., p1_n, ...]``

        Or cell connectivity given as list of node IDs:
        ``[p0_0, p0_1, ..., p0_n, p1_0, p1_1, ..., p1_n, ...]``

    connectivity : bool, optional
        Indicate that cells are given by connectivity. Default: False

    Returns
    -------
    np.ndarray
        Cell nodes as 2D array.
    """
    unique_cell_types = np.unique(cell_types)
    max_nodes = np.max(NODE_COUNT[unique_cell_types])
    cells_matrix = np.full((len(cell_types), max_nodes), fill_value=-1, dtype=int)
    cell_sizes = NODE_COUNT[cell_types]

    if connectivity:
        cell_ends = np.cumsum(cell_sizes)
        cell_starts = np.concatenate(
            [np.array([0], dtype=cell_ends.dtype), cell_ends[:-1]]
        )
    else:
        # adding one to skip cell size entry in cell definitions array
        # [(n0), p0_0, p0_1, ..., p0_n, (n1), p1_0, p1_1, ..., p1_n, ...]
        cell_ends = np.cumsum(cell_sizes + 1)
        cell_starts = (
            np.concatenate([np.array([0], dtype=cell_ends.dtype), cell_ends[:-1]]) + 1
        )

    for cell_type in unique_cell_types:
        cell_size = NODE_COUNT[cell_type]
        mask = cell_types == cell_type
        current_cell_starts = cell_starts[mask]

        cells_inds = current_cell_starts[..., np.newaxis] + np.arange(cell_size)[
            np.newaxis
        ].astype(cell_starts.dtype)
        cells_matrix[:, :cell_size][mask] = cells[cells_inds]

    return cells_matrix


class Location(Enum):
    """Data location in the grid."""

    CELLS = 0
    POINTS = 1


class CellType(IntEnum):
    """Supported cell types."""

    # VTK and ESMF cell node order is counter clockwise
    # https://kitware.github.io/vtk-examples/site/VTKFileFormats/#legacy-file-examples
    # https://earthsystemmodeling.org/docs/release/latest/ESMF_refdoc/node5.html#const:meshelemtype

    # vertex for no-cell meshes
    VERTEX = 0
    # lines for 1D
    LINE = 1
    # 2D following ESMF
    TRI = 2
    QUAD = 3
    # 3D following ESMF
    TETRA = 4
    HEX = 5


NODE_COUNT = np.array([1, 2, 3, 4, 4, 8], dtype=int)
"""np.ndarray: Node numbers per CellType."""


CELL_DIM = np.array([0, 1, 2, 2, 3, 3], dtype=int)
"""np.ndarray: Cell dimension per CellType."""


ESMF_TYPE_MAP = np.array([-1, -1, 3, 4, 10, 12], dtype=int)
"""np.ndarray: ESMF type per CellType (-1 for unsupported)."""


VTK_TYPE_MAP = np.array((vtk_t := [1, 3, 5, 9, 10, 12]), dtype=int)
"""np.ndarray: VTK type per CellType."""


INV_VTK_TYPE_MAP = np.array(
    [vtk_t.index(i) if i in vtk_t else -1 for i in range(82)],
    dtype=int,
)
"""np.ndarray: CellType per VTK type."""


# VTK v9.3
VTK_CELL_DIM = np.array(
    # Linear cells (0-16)
    [0, 0, 0, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3] + 4 * [-1]
    # Quadratic / Cubic, isoparametric cells (21-37)
    + [1, 2, 2, 3, 3, 3, 3, 2, 3, 2, 3, 3, 3, 2, 1, 2, 3] + 3 * [-1]
    # Special class and Polyhedron cell (41-42)
    + [1, 3] + 8 * [-1]
    # Higher order cells in parametric form (51-56)
    + [1, 2, 2, 2, 3, 3] + 3 * [-1]
    # Higher order cells (60-67)
    + [1, 2, 2, 2, 3, 3, 3, 3]
    # Arbitrary order Lagrange elements (68-74)
    + [1, 2, 2, 3, 3, 3, 3]
    # Arbitrary order Bezier elements (75-81)
    + [1, 2, 2, 3, 3, 3, 3],
    dtype=int,
)
"""np.ndarray: Cell dimension per VTK type."""



"""
FINAM-specific error types.

Errors
======

.. autosummary::
   :toctree: generated

    :noindex: FinamDataError
    :noindex: FinamLogError
    :noindex: FinamMetaDataError
    :noindex: FinamNoDataError
    :noindex: FinamStatusError
    :noindex: FinamTimeError
"""


class FinamStatusError(Exception):
    """Error for unexpected status in Components."""


class FinamTimeError(Exception):
    """Error for request time not matching available data timestamps."""


class FinamLogError(Exception):
    """Error for wrong logging configuration."""


class FinamNoDataError(Exception):
    """Error for data not yet being available."""


class FinamMetaDataError(Exception):
    """Error for missing but required metadata."""


class FinamDataError(Exception):
    """Error for wrong data."""


class FinamStaticDataError(Exception):
    """Error for operations that are not allowed on static inputs or outputs."""


class FinamConnectError(Exception):
    """Error for wrong connection setup."""


class FinamCircularCouplingError(Exception):
    """Error for unresolved circular or bidirectional coupling."""



"""Components for controlling data flow"""
import datetime as dt

from ..data.grid_spec import NoGrid
from ..errors import FinamMetaDataError
from ..sdk import TimeComponent
from ..tools.connect_helper import FromInput, FromOutput
from ..tools.log_helper import ErrorLogger


class TimeTrigger(TimeComponent):
    """Component to forward data in regular time intervals.

    Can be used to enable coupling of pull-based source and push-based target components,
    that would not work due to the dead link otherwise.

    .. code-block:: text

                 +-------------+
                 |             |
        --> [In] | TimeTrigger | [Out] -->
                 |             |
                 +-------------+

    The typical usage is shown in the example below.
    Here, the component will retrieve metadata from it's input,
    for all metadata fields that are ```None``.
    Because ``out_info`` is ``None``, the received metadata is then forwarded to the output.

    The same mechanism can be used in the other direction, by specifying ``out_info``
    but not ``in_info`` in the method call.

    Examples
    --------

    .. testcode:: constructor

        import datetime as dt
        import finam as fm

        component = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=dt.datetime(2000, 1, 1),
            step=dt.timedelta(days=1),
        )

    .. testcode:: constructor
        :hide:

        component.initialize()

    Parameters
    ----------
    start : :class:`datetime <datetime.datetime>`
        Starting time. Can be ``None`` to retrieve it from linked components.
        See parameter ``start_from_input`` for details.
    step : :class:`timedelta <datetime.timedelta>` or :class:`relativedelta <dateutil.relativedelta.relativedelta>`
        The component's time step
    in_info : :class:`.Info`, optional
        Input info, optional. However, one of ``in_info`` or ``out_info`` must be given.
        ``time`` is ignored and can be set to ``None``.
    out_info : :class:`.Info`, optional
        Output info, optional. However, one of ``in_info`` or ``out_info`` must be given.
        ``time`` is ignored and can be set to ``None``.
    start_from_input : bool, optional
        Whether to get the starting time from the input, instead of the output. Default ``True``.

        If ``start`` is ``None``, the component can try to retrieve the starting time either
        from the input or from the output.
        The respective linked component should have an internal time step.
        If both linked components have no time step, ``start`` must be given.
    """

    def __init__(self, start, step, in_info=None, out_info=None, start_from_input=True):
        super().__init__()

        self._ini_in_info = in_info
        self._ini_out_info = out_info

        self._start = start
        if self._start is not None:
            self.time = self._start

        self._step = step
        self._start_from_input = start_from_input

    def _next_time(self):
        return self.time + self._step

    def _initialize(self):
        with ErrorLogger(self.logger):
            if self._ini_in_info is None and self._ini_out_info is None:
                raise FinamMetaDataError(
                    "At least one of input or output info must be given."
                )
            if self._start is None:
                if self._start_from_input and self._ini_in_info is None:
                    raise FinamMetaDataError(
                        "Can't get starting time from the input without an input info."
                    )
                if not self._start_from_input and self._ini_out_info is None:
                    raise FinamMetaDataError(
                        "Can't get starting time from the output without an output info."
                    )

        if self._ini_in_info is not None:
            self._ini_in_info.time = self._start
        if self._ini_out_info is not None:
            self._ini_out_info.time = self._start

        in_info_rules = {}
        out_info_rules = {}

        if self._start is None:
            if self._start_from_input:
                self.inputs.add(name="In", info=self._ini_in_info)
                self.outputs.add(name="Out")
                out_info_rules["Out"] = [FromInput("In")]
            else:
                self.inputs.add(name="In")
                self.outputs.add(name="Out", info=self._ini_out_info)
                in_info_rules["In"] = [FromOutput("Out")]
        else:
            self.inputs.add(name="In", info=self._ini_in_info)
            self.outputs.add(name="Out", info=self._ini_out_info)
            if self._ini_out_info is None:
                out_info_rules["Out"] = [FromInput("In")]
            if self._ini_in_info is None:
                in_info_rules["In"] = [FromOutput("Out")]

        self.create_connector(
            pull_data=["In"],
            in_info_rules=in_info_rules,
            out_info_rules=out_info_rules,
        )

    def _connect(self, start_time):
        out_data = {}
        if (
            not self.connector.data_pushed["Out"]
            and self.connector.in_data["In"] is not None
        ):
            out_data["Out"] = self.connector.in_data["In"]

        self.try_connect(start_time, push_data=out_data)

        in_info = self.connector.in_infos["In"]
        if in_info is not None:
            self.time = in_info.time

    def _validate(self):
        pass

    def _update(self):
        self.time += self._step

        data = self.inputs["In"].pull_data(self.time)
        self.outputs["Out"].push_data(data, self.time)

    def _finalize(self):
        pass


class UserControl(TimeComponent):

    """Component to allow users to step a simulation.

    Prompts for input on the console.

    Users can just press ENTER to proceed by one step, or enter a target time in ISO format.

    .. code-block:: text

                 +-------------+
                 |             |
                 | UserControl | [Counter] -->
                 |             |
                 +-------------+

    Examples
    --------

    .. testcode:: constructor

        import datetime as dt
        import finam as fm

        component = fm.components.UserControl(
            start=dt.datetime(2000, 1, 1),
            step=dt.timedelta(days=1),
        )

    .. testcode:: constructor
        :hide:

        component.initialize()

    .. |relativedelta| replace:: :class:`relativedelta <dateutil.relativedelta.relativedelta>`

    Parameters
    ----------
    start : :class:`datetime <datetime.datetime>`
        Starting time. Can be ``None`` to retrieve it from linked components.
    step : :class:`timedelta <datetime.timedelta>` or |relativedelta|, optional
        The component's time step. Default 1 day.
    """

    def __init__(self, start, step=dt.timedelta(days=1)):
        super().__init__()

        self.time = start
        self.step = step
        self._counter = 0

    def _next_time(self):
        return None

    def _initialize(self):
        self.outputs.add(name="Counter", time=self.time, grid=NoGrid())

        self.create_connector()

    def _connect(self, start_time):
        push_data = {}
        if self.connector.data_required["Counter"]:
            push_data["Counter"] = self._counter

        self.try_connect(start_time, push_data=push_data)

    def _validate(self):
        pass

    def _update(self):
        self._counter += 1
        self._prompt()

    def _prompt(self):
        run_until = None

        inp = input(f"Time: {self.time} - Run until (ENTER to step): ")
        if inp == "":
            run_until = self.time + self.step
        else:
            try:
                run_until = dt.datetime.fromisoformat(inp)
            except ValueError:
                print(f"Not a time: '{inp}'.")

        if run_until is not None:
            self.time = run_until
            self.outputs["Counter"].push_data(self._counter, self.time)

    def _finalize(self):
        pass



"""Noise generator components"""
import datetime as dt

import numpy as np
import opensimplex as ox

from finam.data.grid_base import Grid
from finam.data.grid_spec import NoGrid, UnstructuredGrid
from finam.data.tools import Info
from finam.errors import FinamMetaDataError
from finam.interfaces import ComponentStatus
from finam.sdk import CallbackOutput, Component, Output
from finam.tools import ErrorLogger


class SimplexNoise(Component):
    """Pull-based simplex noise generator.

    Requires `opensimplex <https://pypi.org/project/opensimplex/>`_ to be installed.

    .. code-block:: text

        +--------------+
        |              |
        | SimplexNoise | [Noise] -->
        |              |
        +--------------+

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        component = fm.components.SimplexNoise(
            info=fm.Info(time=None, grid=fm.UniformGrid((20, 15))),
            frequency=0.1,
            time_frequency=1 / (24 * 3600),
            octaves=3,
            persistence=0.75,
            low=0.0, high=1.0,
            seed=1234,
        )

    .. testcode:: constructor
        :hide:

        component.initialize()

    Parameters
    ----------
    info : Info, optional
        Output metadata info. All values are taken from the target if not specified.
    frequency : float
        Spatial frequency of the noise, in map units. Default 1.0
    time_frequency : float
        Temporal frequency of the noise, in Hz. Default 1.0
    octaves : int
        Number of octaves. Default 1
    persistence : float, optional
        Persistence over octaves. Default 0.5
    low : float, optional
        Lower limit for values. Default -1.0
    high : float, optional
        Upper limit for values. Default 1.0
    seed : int, optional
        PRNG seed for noise generator. Default 0
    """

    def __init__(
        self,
        info=None,
        frequency=1.0,
        time_frequency=1.0,
        octaves=1,
        persistence=0.5,
        low=-1,
        high=1,
        seed=0,
    ):
        super().__init__()

        if octaves < 1:
            raise ValueError("At least one octave required for SimplexNoise")

        self._info = info or Info(time=None, grid=None, units=None)
        self._seed = seed

        self._frequency = frequency
        self._time_frequency = time_frequency
        self._persistence = persistence
        self._low = low
        self._high = high
        self._octaves = octaves

        self._is_ready = False

    def _initialize(self):
        self.outputs.add(
            io=CallbackOutput(
                callback=self._generate_noise, name="Noise", info=self._info
            )
        )
        self.create_connector()

    def _connect(self, start_time):
        self.try_connect(start_time)

        info = self.connector.out_infos["Noise"]
        if info is not None:
            failed = isinstance(info.grid, NoGrid) and info.grid.dim > 0
            failed |= isinstance(info.grid, Grid) and not 1 <= info.grid.dim <= 3
            if failed:
                with ErrorLogger(self.logger):
                    raise FinamMetaDataError(
                        f"Can generate simplex noise only for gridded 1D-3D data, or for 0D 'NoGrid' data. "
                        f"Got {info.grid}"
                    )

            self._info = info
            self._is_ready = True

    def _validate(self):
        pass

    def _update(self):
        pass

    def _finalize(self):
        pass

    def _generate_noise(self, _caller, time):
        if not self._is_ready:
            return None

        ox.seed(self._seed)

        grid = self._info.grid
        t = (time - dt.datetime(1900, 1, 1)).total_seconds()

        return _generate_noise(
            grid,
            t,
            self._frequency,
            self._time_frequency,
            self._octaves,
            self._persistence,
            self._low,
            self._high,
        )


class StaticSimplexNoise(Component):
    """Static simplex noise generator.

    Requires `opensimplex <https://pypi.org/project/opensimplex/>`_ to be installed.

    .. code-block:: text

        +--------------------+
        |                    |
        | StaticSimplexNoise | [Noise] -->
        |                    |
        +--------------------+

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        component = fm.components.StaticSimplexNoise(
            info=fm.Info(time=None, grid=fm.UniformGrid((20, 15))),
            frequency=0.1,
            octaves=3,
            persistence=0.75,
            low=0.0, high=1.0,
            seed=1234,
        )

    .. testcode:: constructor
        :hide:

        component.initialize()

    Parameters
    ----------
    info : Info, optional
        Output metadata info. All values are taken from the target if not specified.
    frequency : float
        Spatial frequency of the noise, in map units. Default 1.0
    octaves : int
        Number of octaves. Default 1
    persistence : float, optional
        Persistence over octaves. Default 0.5
    low : float, optional
        Lower limit for values. Default -1.0
    high : float, optional
        Upper limit for values. Default 1.0
    seed : int, optional
        PRNG seed for noise generator. Default 0
    """

    def __init__(
        self,
        info=None,
        frequency=1.0,
        octaves=1,
        persistence=0.5,
        low=-1,
        high=1,
        seed=0,
    ):
        super().__init__()

        if octaves < 1:
            raise ValueError("At least one octave required for SimplexNoise")

        self._info = info or Info(time=None, grid=None, units=None)
        self._seed = seed

        self._frequency = frequency
        self._persistence = persistence
        self._low = low
        self._high = high
        self._octaves = octaves

        self._is_ready = False

    def _initialize(self):
        self.outputs.add(io=Output(name="Noise", static=True, info=self._info))
        self.create_connector()

    def _connect(self, start_time):
        push_data = {}
        if self.connector.out_infos["Noise"] is not None:
            push_data["Noise"] = self._generate_noise()

        self.try_connect(start_time, push_data=push_data)

        info = self.connector.out_infos["Noise"]
        if info is not None:
            failed = isinstance(info.grid, NoGrid) and info.grid.dim > 0
            failed |= isinstance(info.grid, Grid) and not 1 <= info.grid.dim <= 3
            if failed:
                with ErrorLogger(self.logger):
                    raise FinamMetaDataError(
                        f"Can generate simplex noise only for gridded 1D-3D data, or for 0D 'NoGrid' data. "
                        f"Got {info.grid}"
                    )

            self._info = info
            self._is_ready = True

    def _validate(self):
        self.status = ComponentStatus.FINISHED

    def _update(self):
        pass

    def _finalize(self):
        pass

    def _generate_noise(self):
        ox.seed(self._seed)

        grid = self._info.grid

        return _generate_noise(
            grid,
            1,
            self._frequency,
            1,
            self._octaves,
            self._persistence,
            self._low,
            self._high,
        )


def _generate_noise(
    grid, time, frequency, time_frequency, octaves, persistence, low, high
):
    if isinstance(grid, NoGrid):
        func = _generate_scalar
    elif isinstance(grid, UnstructuredGrid):
        func = _generate_unstructured
    else:
        func = _generate_structured

    amp = 1.0
    max_amp = 0.0
    freq = frequency
    freq_t = time_frequency

    for i in range(octaves):
        if i == 0:
            data = func(grid, time * freq_t, freq)
        else:
            data += amp * func(grid, time * freq_t, freq)
        max_amp += amp
        amp *= persistence
        freq *= 2.0
        freq_t *= 2.0

    data /= max_amp
    data = data * (high - low) / 2 + (high + low) / 2
    return data


def _generate_scalar(_grid, t, _freq):
    return ox.noise2(0, t)


def _generate_structured(grid, t, freq):
    if grid.dim == 1:
        data = ox.noise2array(grid.data_axes[0] * freq, np.asarray([t]))
    if grid.dim == 2:
        data = ox.noise3array(
            grid.data_axes[0] * freq, grid.data_axes[1] * freq, np.asarray([t])
        )
    if grid.dim == 3:
        data = ox.noise4array(
            grid.data_axes[0] * freq,
            grid.data_axes[1] * freq,
            grid.data_axes[2] * freq,
            np.asarray([t]),
        )
    # pylint: disable-next=possibly-used-before-assignment
    return data[0, ...].T


def _generate_unstructured(grid, t, freq):
    points = grid.data_points
    data = np.full((points.shape[0],), 0.0, dtype=float)

    if grid.dim == 1:
        for i, p in enumerate(points):
            data[i] = ox.noise2(p[0] * freq, t)
    elif grid.dim == 2:
        for i, p in enumerate(points):
            data[i] = ox.noise3(p[0] * freq, p[1] * freq, t)
    elif grid.dim == 3:
        for i, p in enumerate(points):
            data[i] = ox.noise4(p[0] * freq, p[1] * freq, p[2] * freq, t)

    return data



"""
Implementations of IInput
"""
import logging
from datetime import datetime

from ..data import tools
from ..data.tools import Info
from ..errors import FinamMetaDataError
from ..interfaces import IInput, IOutput, Loggable
from ..tools.log_helper import ErrorLogger


class Input(IInput, Loggable):
    """Default input implementation."""

    def __init__(self, name, info=None, static=False, **info_kwargs):
        Loggable.__init__(self)
        self._source = None
        self.base_logger_name = None
        if name is None:
            raise ValueError("Input: needs a name.")
        self._name = name
        self._static = static
        if info_kwargs:
            if info is not None:
                raise ValueError("Input: can't use **kwargs in combination with info")
            info = Info(**info_kwargs)
        self._input_info = info
        self._in_info_exchanged = False
        self._cached_data = None
        self._transform = None

    @property
    def name(self):
        """Input name."""
        return self._name

    @property
    def is_static(self):
        return self._static

    @property
    def info(self):
        """Info: The input's data info."""
        return self._input_info

    @property
    def needs_pull(self):
        """bool: if the input needs pull."""
        return True

    @property
    def needs_push(self):
        """bool: if the input needs push."""
        return False

    @property
    def source(self):
        """Get the input's source output or adapter

        Returns
        -------
        :class:`.IOutput`
            The input's source.
        """
        return self._source

    @source.setter
    def source(self, source):
        """Set the input's source output or adapter

        Parameters
        ----------
        source : :class:`.IOutput`
            source output or adapter
        """
        self.logger.trace("set source")

        with ErrorLogger(self.logger):
            if self._source is not None:
                raise ValueError(
                    "Source of input is already set! "
                    "(You probably tried to connect multiple outputs to a single input)"
                )
            if not isinstance(source, IOutput):
                raise ValueError("Only IOutput can be set as source for Input")

        self._source = source

    def source_updated(self, time):
        """Informs the input that a new output is available.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the notification.
        """
        self.logger.trace("source changed")

    def pull_data(self, time, target=None):
        """Retrieve the data from the input's source.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time to get the data for.
        target : :class:`.IInput` or None
            Requesting end point of this pull.
            Should be ``None`` for normal input pulls in components.
            Simple adapters should forward the source in :meth:`.Adapter._get_data`.
            Push-based adapters should use ``self`` in :meth:`.Adapter._source_updated`.

        Returns
        -------
        :class:`pint.Quantity`
            Data set for the given simulation time.
        """
        self.logger.trace("pull data")

        if time is not None and not isinstance(time, datetime):
            with ErrorLogger(self.logger):
                raise ValueError("Time must be of type datetime")

        if self.is_static:
            if self._cached_data is None:
                data = self._source.get_data(time, target or self)
                with ErrorLogger(self.logger):
                    self._cached_data = self._convert_and_check(data)
            data = self._cached_data
        else:
            data = self._source.get_data(time, target or self)
            with ErrorLogger(self.logger):
                data = self._convert_and_check(data)

        return data

    def _convert_and_check(self, data):
        # transform compatible data between grids
        if self._transform is not None:
            with ErrorLogger(self.logger):
                data = self._transform(data)
            self.logger.profile(
                "converted data between compatible grids (%d entries)", data.size
            )

        # convert units
        data, conv = tools.to_units(
            data, self._input_info.units, check_equivalent=True, report_conversion=True
        )
        if conv is not None:
            self.logger.profile(
                "converted units from %s to %s (%d entries)", *conv, data.size
            )
        tools.check(data, self._input_info)
        return data

    def ping(self):
        """Pings upstream to inform outputs about the number of connected inputs.

        Must be called after linking and before the connect phase.
        """
        self._source.pinged(self)

    def exchange_info(self, info=None):
        """Exchange the data info with the input's source.

        Parameters
        ----------
        info : :class:`.Info`
            request parameters

        Returns
        -------
        dict
            delivered parameters
        """
        self.logger.trace("exchanging info")

        with ErrorLogger(self.logger):
            if self._in_info_exchanged:
                raise FinamMetaDataError("Input info was already exchanged.")
            if self._input_info is not None and info is not None:
                raise FinamMetaDataError("An internal info was already provided")
            if self._input_info is None and info is None:
                raise FinamMetaDataError("No metadata provided")
            if info is None:
                info = self._input_info

            if not isinstance(info, Info):
                raise FinamMetaDataError("Metadata must be of type Info")

        src_info = self._source.get_info(info)

        with ErrorLogger(self.logger):
            fail_info = {}
            if not info.accepts(src_info, fail_info):
                fail_info = "\n".join(
                    [
                        f"{name} - got {got}, expected {exp}"
                        for name, (got, exp) in fail_info.items()
                    ]
                )
                raise FinamMetaDataError(
                    f"Can't accept incoming data info. Failed entries:\n{fail_info}"
                )

        self._input_info = src_info.copy_with(
            use_none=False, time=info.time, grid=info.grid, **info.meta
        )
        self._in_info_exchanged = True
        with ErrorLogger(self.logger):
            self._transform = src_info.grid.get_transform_to(self._input_info.grid)

        # pylint: disable-next=fixme
        # TODO: check if this is correct (was src_info before)
        return self._input_info

    @property
    def has_source(self):
        """Flag if this input instance has a source."""
        return self._source is not None

    @property
    def logger_name(self):
        """Logger name derived from base logger name and class name."""
        base_logger = logging.getLogger(self.base_logger_name)
        # logger hierarchy indicated by "." in name
        return ".".join(([base_logger.name, "<-", self.name]))

    @property
    def uses_base_logger_name(self):
        """Whether this class has a ``base_logger_name`` attribute. True."""
        return True


class CallbackInput(Input):
    """Input implementation calling a callback when notified.

    Use for components without time step.

    Parameters
    ----------
    callback : callable
        A callback ``callback(caller, time)``, returning the transformed data.
    """

    def __init__(self, callback, name, info=None, static=False, **info_kwargs):
        super().__init__(name=name, info=info, static=static, **info_kwargs)
        self.callback = callback

    @property
    def needs_pull(self):
        """bool: if the input needs pull."""
        return False

    @property
    def needs_push(self):
        """bool: if the input needs push."""
        return True

    def source_updated(self, time):
        """Informs the input that a new output is available.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the notification.
        """
        self.logger.trace("source changed")
        if time is not None and not isinstance(time, datetime):
            with ErrorLogger(self.logger):
                raise ValueError("Time must be of type datetime")

        self.callback(self, time)



"""
Unit tests for the adapters.time module.
"""

import unittest
from datetime import datetime, timedelta

import numpy as np
import pint

import finam as fm
from finam import FinamTimeError, Info, NoGrid, UniformGrid
from finam import data as tools
from finam.adapters.time_integration import AvgOverTime, SumOverTime
from finam.components import CallbackGenerator, DebugConsumer

reg = pint.UnitRegistry(force_ndarray_like=True)


class TestAvgOverTime(unittest.TestCase):
    def init(self, step):
        start = datetime(2000, 1, 1)

        self.source = CallbackGenerator(
            callbacks={
                "Step": (lambda t: t.day - 1, Info(None, grid=NoGrid(), units="m"))
            },
            start=start,
            step=timedelta(1.0),
        )

        self.adapter = AvgOverTime(step=step)

        self.source.initialize()

        self.source.outputs["Step"] >> self.adapter
        self.adapter.get_info(Info(None, grid=NoGrid(), units=None))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_avg_over_time(self):
        self.init(None)

        self.source.update()
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 1, 12), None)),
            0.25,
        )
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 2, 0), None)),
            0.75,
        )
        self.source.update()
        self.source.update()
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 4, 0), None)),
            2.0,
        )

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 1, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 5, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(100, None)

    def test_avg_step_over_time(self):
        self.init(0.25)

        self.source.update()
        self.source.update()

        data = self.adapter.get_data(datetime(2000, 1, 1, 4), None)
        self.assertAlmostEqual(tools.get_magnitude(data).item(), 0.0)

        data = self.adapter.get_data(datetime(2000, 1, 1, 7), None)
        self.assertAlmostEqual(tools.get_magnitude(data).item(), 1 / 3)

        data = self.adapter.get_data(datetime(2000, 1, 1, 20), None)
        self.assertAlmostEqual(tools.get_magnitude(data).item(), 1)

        data = self.adapter.get_data(datetime(2000, 1, 2, 4), None)
        self.assertAlmostEqual(tools.get_magnitude(data).item(), 1)

        data = self.adapter.get_data(datetime(2000, 1, 2, 8), None)
        self.assertAlmostEqual(tools.get_magnitude(data).item(), 1.5)


class TestGridAvgOverTime(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)

        grid, _ = create_grid(10, 15, 0)
        self.source = CallbackGenerator(
            callbacks={
                "Grid": (
                    lambda t: create_grid(10, 15, t.day - 1)[1],
                    Info(None, grid=grid),
                )
            },
            start=start,
            step=timedelta(1.0),
        )

        self.adapter = AvgOverTime()

        self.source.initialize()

        self.source.outputs["Grid"] >> self.adapter
        self.adapter.get_info(Info(None, grid=grid, units=None))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_linear_grid_integration(self):
        self.source.update()
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 1, 12), None))[
                0, 2, 3
            ],
            0.25,
        )
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 2, 0), None))[
                0, 2, 3
            ],
            0.75,
        )
        self.source.update()
        self.source.update()
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 4, 0), None))[
                0, 2, 3
            ],
            2.0,
        )

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 1, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 5, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(100, None)


class TestSumOverTime(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)

        self.data = {
            "In_lin": [],
            "In_00": [],
            "In_05": [],
            "In_10": [],
        }

        def callback(n, d, t):
            self.data[n].append((t, d))

        self.source = CallbackGenerator(
            callbacks={
                "Step": (
                    lambda t: t.day * fm.UNITS.Unit("mm/d"),
                    Info(None, grid=NoGrid(), units="mm/d"),
                )
            },
            start=start,
            step=timedelta(days=1),
        )
        self.adapter_lin = SumOverTime(step=None)
        self.adapter_00 = SumOverTime(step=0.0)
        self.adapter_05 = SumOverTime(step=0.5)
        self.adapter_10 = SumOverTime(step=1.0)

        self.consumer = DebugConsumer(
            inputs={
                "In_lin": fm.Info(time=None, grid=NoGrid(), units="mm"),
                "In_00": fm.Info(time=None, grid=NoGrid(), units="mm"),
                "In_05": fm.Info(time=None, grid=NoGrid(), units="mm"),
                "In_10": fm.Info(time=None, grid=NoGrid(), units="mm"),
            },
            callbacks={
                "In_lin": callback,
                "In_00": callback,
                "In_05": callback,
                "In_10": callback,
            },
            start=start,
            step=timedelta(days=2),
        )

        self.comp = fm.Composition([self.source, self.consumer])

        self.source["Step"] >> self.adapter_lin >> self.consumer["In_lin"]
        self.source["Step"] >> self.adapter_00 >> self.consumer["In_00"]
        self.source["Step"] >> self.adapter_05 >> self.consumer["In_05"]
        self.source["Step"] >> self.adapter_10 >> self.consumer["In_10"]

    def test_sum_over_time(self):
        self.comp.run(end_time=datetime(2000, 1, 10))

        mm = fm.UNITS.Unit("mm")

        res = [d.item(0) for t, d in self.data["In_lin"]]
        exp = [0 * mm, 4 * mm, 8 * mm, 12 * mm, 16 * mm, 20 * mm]

        for r, e in zip(res, exp):
            self.assertEqual(r.units, e.units)
            self.assertAlmostEqual(r.magnitude, e.magnitude)

        res = [d.item(0) for t, d in self.data["In_05"]]
        exp = [0 * mm, 4 * mm, 8 * mm, 12 * mm, 16 * mm, 20 * mm]

        for r, e in zip(res, exp):
            self.assertEqual(r.units, e.units)
            self.assertAlmostEqual(r.magnitude, e.magnitude)

        res = [d.item(0) for t, d in self.data["In_00"]]
        exp = [0 * mm, 5 * mm, 9 * mm, 13 * mm, 17 * mm, 21 * mm]

        for r, e in zip(res, exp):
            self.assertEqual(r.units, e.units)
            self.assertAlmostEqual(r.magnitude, e.magnitude)

        res = [d.item(0) for t, d in self.data["In_10"]]
        exp = [0 * mm, 3 * mm, 7 * mm, 11 * mm, 15 * mm, 19 * mm]

        for r, e in zip(res, exp):
            self.assertEqual(r.units, e.units)
            self.assertAlmostEqual(r.magnitude, e.magnitude)


class TestAbsSumOverTime(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)

        self.data = {
            "In_lin": [],
            "In_00": [],
            "In_05": [],
            "In_10": [],
        }

        def callback(n, d, t):
            self.data[n].append((t, d))

        self.source = CallbackGenerator(
            callbacks={
                "Step": (
                    lambda t: 0.5 * (t.day + 1) * fm.UNITS.Unit("mm"),
                    Info(None, grid=NoGrid(), units="mm"),
                )
            },
            start=start,
            step=timedelta(days=2),
        )
        self.adapter_lin = SumOverTime(step=None, per_time=False)
        self.adapter_00 = SumOverTime(step=0.0, per_time=False)
        self.adapter_05 = SumOverTime(step=0.5, per_time=False)
        self.adapter_10 = SumOverTime(step=1.0, per_time=False)

        self.consumer = DebugConsumer(
            inputs={
                "In_lin": fm.Info(time=None, grid=NoGrid(), units="mm"),
                "In_00": fm.Info(time=None, grid=NoGrid(), units="mm"),
                "In_05": fm.Info(time=None, grid=NoGrid(), units="mm"),
                "In_10": fm.Info(time=None, grid=NoGrid(), units="mm"),
            },
            callbacks={
                "In_lin": callback,
                "In_00": callback,
                "In_05": callback,
                "In_10": callback,
            },
            start=start,
            step=timedelta(days=4),
        )

        self.comp = fm.Composition([self.source, self.consumer])

        self.source["Step"] >> self.adapter_lin >> self.consumer["In_lin"]
        self.source["Step"] >> self.adapter_00 >> self.consumer["In_00"]
        self.source["Step"] >> self.adapter_05 >> self.consumer["In_05"]
        self.source["Step"] >> self.adapter_10 >> self.consumer["In_10"]

    def test_sum_over_time(self):
        self.comp.run(end_time=datetime(2000, 1, 20))

        mm = fm.UNITS.Unit("mm")

        res = [d.item(0) for t, d in self.data["In_lin"]]
        exp = [1 * mm, 4 * mm, 8 * mm, 12 * mm, 16 * mm, 20 * mm]

        for r, e in zip(res, exp):
            self.assertEqual(r.units, e.units)
            self.assertAlmostEqual(r.magnitude, e.magnitude)

        res = [d.item(0) for t, d in self.data["In_05"]]
        exp = [1 * mm, 4 * mm, 8 * mm, 12 * mm, 16 * mm, 20 * mm]

        for r, e in zip(res, exp):
            self.assertEqual(r.units, e.units)
            self.assertAlmostEqual(r.magnitude, e.magnitude)

        res = [d.item(0) for t, d in self.data["In_00"]]
        exp = [1 * mm, 5 * mm, 9 * mm, 13 * mm, 17 * mm, 21 * mm]

        for r, e in zip(res, exp):
            self.assertEqual(r.units, e.units)
            self.assertAlmostEqual(r.magnitude, e.magnitude)

        res = [d.item(0) for t, d in self.data["In_10"]]
        exp = [1 * mm, 3 * mm, 7 * mm, 11 * mm, 15 * mm, 19 * mm]

        for r, e in zip(res, exp):
            self.assertEqual(r.units, e.units)
            self.assertAlmostEqual(r.magnitude, e.magnitude)


def create_grid(cols, rows, value):
    grid = UniformGrid((cols, rows), data_location="POINTS")

    data = np.full(shape=grid.data_shape, fill_value=value, order=grid.order)

    return grid, data


if __name__ == "__main__":
    unittest.main()



"""Working directory helpers."""
import os
from contextlib import contextmanager

from .log_helper import ErrorLogger, is_loggable


@contextmanager
def set_directory(path):
    """Sets the cwd within a context.

    Parameters
    ----------
    path : Path
        Path to the desired cwd.

    Yields
    ------
    None
    """
    origin = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)


def execute_in_cwd(func):
    """Decorator to execute method in given cwd of containing class.

    Parameters
    ----------
    func : callable
        Method of a class.

    Returns
    -------
    callable
        Wrapped function to be executed in cwd.

    Notes
    -----
    The containing class of the given method needs to have a "cwd" property.
    """

    def cwd_wrapper(self, *args, **kwargs):
        """Wrapper function."""
        cwd = getattr(self, "cwd", None)
        with ErrorLogger(getattr(self, "logger", None), do_log=is_loggable(self)):
            if cwd is None:
                raise ValueError("No working directory given.")
        with set_directory(cwd):
            return func(self, *args, **kwargs)

    return cwd_wrapper



import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

from finam import (
    CellType,
    EsriGrid,
    Location,
    RectilinearGrid,
    UniformGrid,
    UnstructuredGrid,
    UnstructuredPoints,
)

HEADER = [
    "ncols",
    "nrows",
    "xllcorner",
    "yllcorner",
    "xllcenter",
    "yllcenter",
    "cellsize",
    "NODATA_value",
]


def write_asc(file, info):
    with open(file, "w") as f:
        for val in HEADER:
            if val not in info:
                continue
            print(f"{val:<12}", info[val], file=f)


class TestGridSpec(unittest.TestCase):
    def test_uniform(self):
        grid = UniformGrid((3, 2), spacing=(1.0, 2.0), origin=(2.0, 1.0))

        self.assertEqual(grid.name, "UniformGrid")
        self.assertEqual(grid.dim, 2)
        self.assertEqual(grid.dims, (3, 2))
        self.assertEqual(grid.point_count, 6)
        self.assertEqual(grid.cell_count, 2)
        self.assertEqual(grid.data_shape, (2, 1))
        self.assertEqual(grid.axes_names, ["x", "y"])

        assert_allclose(grid.axes[0], [2.0, 3.0, 4.0])
        assert_allclose(grid.axes[1], [1.0, 3.0])

        assert_allclose(grid.cell_axes[0], [2.5, 3.5])
        assert_allclose(grid.cell_axes[1], [2.0])

        assert_allclose(grid.data_axes[0], [2.5, 3.5])
        assert_allclose(grid.data_axes[1], [2.0])

        assert_allclose(
            grid.points,
            [[2.0, 1.0], [3.0, 1.0], [4.0, 1.0], [2.0, 3.0], [3.0, 3.0], [4.0, 3.0]],
        )
        assert_array_equal(grid.cells, [[3, 4, 1, 0], [4, 5, 2, 1]])
        assert_allclose(grid.cell_centers, [[2.5, 2.0], [3.5, 2.0]])

        assert_array_equal(grid.cell_types, [CellType.QUAD, CellType.QUAD])

        with self.assertRaises(ValueError):
            UniformGrid((3, 2), spacing=(1.0,))

        with self.assertRaises(ValueError):
            UniformGrid((3, 2), origin=(0.0,))

        with TemporaryDirectory() as tmp:
            path = Path(tmp) / "test"
            grid.export_vtk(
                path=path, data={"data": np.zeros(grid.data_shape)}, mesh_type="uniform"
            )
            grid.export_vtk(
                path=path,
                data={"data": np.zeros(grid.data_shape)},
                mesh_type="structured",
            )
            grid.export_vtk(
                path=path,
                data={"data": np.zeros(grid.data_shape)},
                mesh_type="unstructured",
            )

        grid = UniformGrid(
            (3, 2), spacing=(1.0, 2.0), origin=(2.0, 1.0), data_location="POINTS"
        )

        with TemporaryDirectory() as tmp:
            path = Path(tmp) / "test"
            grid.export_vtk(
                path=path, data={"data": np.zeros(grid.data_shape)}, mesh_type="uniform"
            )
            grid.export_vtk(
                path=path,
                data={"data": np.zeros(grid.data_shape)},
                mesh_type="structured",
            )
            grid.export_vtk(
                path=path,
                data={"data": np.zeros(grid.data_shape)},
                mesh_type="unstructured",
            )

    def test_rectilinear(self):
        grid = RectilinearGrid([np.asarray([2.0, 3.0, 4.0]), np.asarray([1.0, 3.0])])

        self.assertEqual(grid.name, "RectilinearGrid")
        self.assertIsNone(grid.crs)
        self.assertEqual(grid.dim, 2)
        self.assertEqual(grid.dims, (3, 2))
        self.assertEqual(grid.point_count, 6)
        self.assertEqual(grid.cell_count, 2)
        self.assertEqual(grid.data_shape, (2, 1))
        self.assertEqual(grid.axes_names, ["x", "y"])

        assert_allclose(grid.axes[0], [2.0, 3.0, 4.0])
        assert_allclose(grid.axes[1], [1.0, 3.0])

        assert_allclose(grid.cell_axes[0], [2.5, 3.5])
        assert_allclose(grid.cell_axes[1], [2.0])

        assert_allclose(grid.data_axes[0], [2.5, 3.5])
        assert_allclose(grid.data_axes[1], [2.0])

        assert_allclose(
            grid.points,
            [[2.0, 1.0], [3.0, 1.0], [4.0, 1.0], [2.0, 3.0], [3.0, 3.0], [4.0, 3.0]],
        )
        assert_array_equal(grid.cells, [[3, 4, 1, 0], [4, 5, 2, 1]])
        assert_allclose(grid.cell_centers, [[2.5, 2.0], [3.5, 2.0]])

        assert_array_equal(grid.cell_types, [CellType.QUAD, CellType.QUAD])

        with self.assertRaises(ValueError):
            RectilinearGrid(
                [np.asarray([2.0, 3.0]), np.asarray([1.0, 3.0])], axes_names=["X"]
            )

        with self.assertRaises(ValueError):
            RectilinearGrid(
                [np.asarray([2.0, 3.0]), np.asarray([1.0, 3.0])],
                axes_attributes=[{1: 1}],
            )

        grid = RectilinearGrid([np.asarray([1.0]), np.asarray([2.0, 1.0])])
        assert_array_equal(grid.axes_increase, [True, False])

    def test_unstructured(self):
        grid = UnstructuredGrid(
            points=[
                [0.0, 0.0],
                [0.0, 2.0],
                [2.0, 2.0],
                [2.0, 0.0],
                [2.0, 0.0],
                [2.0, 2.0],
                [4.0, 2.0],
                [4.0, 0.0],
            ],
            cells=[[0, 1, 2, 3], [4, 5, 6, 7]],
            cell_types=[CellType.QUAD, CellType.QUAD],
        )
        grid2 = UnstructuredPoints(
            points=[
                [0.0, 0.0],
                [0.0, 2.0],
                [2.0, 2.0],
                [2.0, 0.0],
                [2.0, 0.0],
                [2.0, 2.0],
                [4.0, 2.0],
                [4.0, 0.0],
            ],
        )

        with self.assertRaises(ValueError):
            UnstructuredGrid(
                points=[
                    [0.0, 0.0],
                    [0.0, 2.0],
                    [2.0, 2.0],
                    [2.0, 0.0],
                    [2.0, 0.0],
                    [2.0, 2.0],
                    [4.0, 2.0],
                    [4.0, 0.0],
                ],
                cells=[[0, 1, 2, 3], [4, 5, 6, 7]],
                cell_types=[CellType.QUAD, CellType.QUAD],
                axes_names=["to_few"],
            )

        with self.assertRaises(ValueError):
            UnstructuredPoints(points=[[0.0, 0.0]], axes_attributes=[{"too": "short"}])

        self.assertEqual(grid.name, "UnstructuredGrid")
        self.assertEqual(grid2.name, "UnstructuredPoints")
        self.assertIsNone(grid.crs)
        self.assertEqual(grid.dim, 2)
        self.assertEqual(grid.order, "C")
        self.assertEqual(grid.mesh_dim, 2)
        self.assertEqual(grid.point_count, 8)
        self.assertEqual(grid.cell_count, 2)
        self.assertEqual(grid.data_shape, (2,))
        self.assertEqual(grid.data_size, 2)
        self.assertEqual(grid.axes_names, ["x", "y"])

        assert_array_equal(grid.cell_types, [3, 3])
        assert_allclose(grid.cell_centers, [[1.0, 1.0], [3.0, 1.0]])
        assert_array_equal(grid.cell_node_counts, [4, 4])
        assert_allclose(grid.data_points, [[1.0, 1.0], [3.0, 1.0]])

        assert_allclose(grid.points, grid2.points)
        assert_allclose(grid.points, grid2.cell_centers)
        self.assertEqual(grid2.dim, 2)
        self.assertEqual(grid2.mesh_dim, 0)
        assert_allclose(grid2.cell_node_counts, 1)

        with self.assertRaises(ValueError):
            grid2.data_location = Location.CELLS

    def test_esri(self):
        header = {
            "ncols": 520,
            "nrows": 600,
            "xllcenter": 4375050.0,
            "yllcenter": 2700050.0,
            "cellsize": 100.0,
        }
        with TemporaryDirectory() as tmp:
            path = Path(tmp) / "test.txt"
            write_asc(path, header)
            grid = EsriGrid.from_file(path)
        self.assertEqual(grid.name, "EsriGrid")
        self.assertAlmostEqual(grid.ncols, header["ncols"])
        self.assertAlmostEqual(grid.nrows, header["nrows"])
        self.assertAlmostEqual(grid.cellsize, header["cellsize"])
        self.assertAlmostEqual(grid.xllcorner, 4375000)
        self.assertAlmostEqual(grid.yllcorner, 2700000)

        # casting
        grid2 = grid.to_uniform()
        grid3 = grid.to_rectilinear()
        grid4 = grid2.to_rectilinear()
        self.assertTrue(grid == grid2)
        self.assertTrue(grid == grid3)
        self.assertTrue(grid == grid4)
        self.assertTrue(grid2 == grid3)
        self.assertTrue(grid2 == grid4)
        self.assertTrue(grid3 == grid4)

        with self.assertRaises(ValueError):
            grid.data_location = Location.POINTS

    def test_data_location(self):
        grid1 = UniformGrid((1,), data_location=0)
        grid2 = UniformGrid((2, 2), data_location="CELLS")
        grid3 = UniformGrid((2, 2), data_location=Location.CELLS)
        grid4 = UniformGrid((2, 2), data_location=1)
        grid5 = UniformGrid((2, 2), data_location="POINTS")
        grid6 = UniformGrid((2, 2), data_location=Location.POINTS)
        self.assertEqual(grid1.data_location, Location.CELLS)
        self.assertEqual(grid2.data_location, Location.CELLS)
        self.assertEqual(grid3.data_location, Location.CELLS)
        self.assertEqual(grid4.data_location, Location.POINTS)
        self.assertEqual(grid5.data_location, Location.POINTS)
        self.assertEqual(grid6.data_location, Location.POINTS)

    def test_cast(self):
        grid = EsriGrid(3, 2)
        us_grid = grid.to_unstructured()
        assert_allclose(grid.data_points, us_grid.data_points)
        self.assertIsInstance(us_grid, UnstructuredGrid)

    def test_copy(self):
        grid = EsriGrid(3, 2)
        us_grid = grid.to_unstructured()
        cp_grid1 = us_grid.copy()
        cp_grid2 = us_grid.copy(deep=True)

        self.assertTrue(us_grid == cp_grid1)
        self.assertTrue(us_grid == cp_grid2)

        # shallow copy shares info
        cp_grid1.points[0, 0] = 0.1
        self.assertTrue(us_grid == cp_grid1)
        self.assertFalse(us_grid == cp_grid2)

    def test_location_check(self):
        grid_s1 = UniformGrid((2, 2), data_location="CELLS")
        grid_s2 = UniformGrid((2, 2), data_location="POINTS")
        grid_u1 = grid_s1.to_unstructured()
        grid_u2 = grid_s2.to_unstructured()
        self.assertTrue(grid_s1.compatible_with(grid_s2, check_location=False))
        self.assertFalse(grid_s1.compatible_with(grid_s2, check_location=True))
        self.assertTrue(grid_u1.compatible_with(grid_u2, check_location=False))
        self.assertFalse(grid_u1.compatible_with(grid_u2, check_location=True))

    def test_equality(self):
        grid1 = UniformGrid((2, 2), data_location=0)
        grid2 = UnstructuredGrid(
            points=[
                [0.0, 0.0],
                [0.0, 2.0],
                [2.0, 2.0],
                [2.0, 0.0],
                [2.0, 0.0],
                [2.0, 2.0],
                [4.0, 2.0],
                [4.0, 0.0],
            ],
            cells=[[0, 1, 2, 3], [4, 5, 6, 7]],
            cell_types=[CellType.QUAD, CellType.QUAD],
            axes_names=["x", "y"],
        )

        self.assertNotEqual(grid1, 0)
        self.assertNotEqual(grid1, grid2)

    def test_cell_types(self):
        grid = UniformGrid((0,))
        self.assertEqual(grid.cell_types, [CellType.VERTEX])

        grid = UniformGrid((1,))
        assert_array_equal(grid.cell_types, [CellType.VERTEX])

        grid = UniformGrid((2,))
        assert_array_equal(grid.cell_types, [CellType.LINE])

        grid = UniformGrid((3,))
        assert_array_equal(grid.cell_types, [CellType.LINE, CellType.LINE])

        grid = UniformGrid((2, 2))
        assert_array_equal(grid.cell_types, [CellType.QUAD])
        grid = UniformGrid((2, 3))
        assert_array_equal(grid.cell_types, [CellType.QUAD, CellType.QUAD])

        grid = UniformGrid((2, 2, 2))
        assert_array_equal(grid.cell_types, [CellType.HEX])
        grid = UniformGrid((2, 2, 3))
        assert_array_equal(grid.cell_types, [CellType.HEX, CellType.HEX])


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime, timedelta
from os import path
from tempfile import TemporaryDirectory

from numpy.testing import assert_array_equal

from finam import Composition, Info, NoGrid
from finam.components.generators import CallbackGenerator
from finam.components.writers import CsvWriter


class TestCsvWriter(unittest.TestCase):
    def test_write_file(self):
        import pandas as pd

        with TemporaryDirectory() as tmp:
            file_path = path.join(tmp, "test.csv")
            start = datetime(2000, 1, 1)

            generator = CallbackGenerator(
                callbacks={
                    "A": (lambda t: 0, Info(None, grid=NoGrid())),
                    "B": (lambda t: (t - start).days, Info(None, grid=NoGrid())),
                    "C": (lambda t: (t - start).days * 2, Info(None, grid=NoGrid())),
                },
                start=start,
                step=timedelta(days=1),
            )

            writer = CsvWriter(
                inputs=["A", "B", "C"],
                path=file_path,
                start=start,
                step=timedelta(days=1),
                separator=",",
            )

            comp = Composition([generator, writer])

            generator.outputs["A"] >> writer.inputs["A"]
            generator.outputs["B"] >> writer.inputs["B"]
            generator.outputs["C"] >> writer.inputs["C"]

            comp.run(start_time=start, end_time=datetime(2000, 1, 31))

            csv = pd.read_csv(file_path)

            assert_array_equal(csv.columns, ["time", "A", "B", "C"])
            assert_array_equal(csv["B"], list(range(0, 31)))
            self.assertEqual(csv.shape[0], 31)

    def test_constructor_fail(self):
        with self.assertRaises(ValueError):
            _writer = CsvWriter(
                inputs=["A", "B", "C"],
                path="abc",
                start=0,
                step=timedelta(days=1),
                separator=",",
            )

        with self.assertRaises(ValueError):
            _writer = CsvWriter(
                inputs=["A", "B", "C"],
                path="abc",
                start=datetime(2000, 1, 1),
                step=1,
                separator=",",
            )


if __name__ == "__main__":
    unittest.main()
